
#include "define.h"
int cond;
int immd;
int rs1;
int rs2;
int rs3;
int rs4;
int rd;
int rd2;
int OP;
int OP2;
int msize;
int msex;
int mwback;
int bit_pos;
int nzcv;
int post_index;
int imask;
int v64;

int d_N;
int d_S;
int d_R;
unsigned long long int tmask;
unsigned long long int wmask;

int inst_reg;

static int opcode;
static int opcode2;
static int op;
static int op1;
static int op2;
static int op3;
static int op4;
static int op5;

static int DecodeRegExtend(int op)
{
  switch(op){
  case 0:
    return UXTB;
  case 1:
    return UXTH;
  case 2:
    return UXTW;
  case 3:
    return UXTX;
  case 4:
    return SXTB;
  case 5:
    return SXTH;
  case 6:
    return SXTW;
  case 7:
    return SXTX;
  }
  return UXTX;
}

static long long int ShiftReg(int m, int ty, int sa)
{
  unsigned long long int valh,vall;
  vall = get_reg(m);
  switch(ty){
  case 0:/* LSL */
    vall = vall<<sa;
    valh = 0ULL;
    break;
  case 1:/* LSR */
    if(v64==0){
      vall &= 0xffffffffULL;
      valh = 0ULL;
    }
    valh = valh <<(64-sa);
    vall = vall>>sa;
    break;
  case 2:/* ASR */
    if(v64==0){
      vall = (vall&(1ULL<<31))? 0xffffffff00000000ULL|vall : 0xffffffffULL&vall;
      valh = 0ULL;
    }
    else{
      valh = (vall&(1ULL<<63))? 0xffffffffffffffffULL : 0ULL;
    }
    valh = valh <<(64-sa);
    vall = vall>>sa;
    break;
  case 3:/* ROR */
    if(v64==0){
      vall = (vall<<32)|(vall&0xffffffffULL);
      valh = 0ULL;
    }
    else{
      valh = vall;
    }
    valh = valh <<(64-sa);
    vall = vall>>sa;
    break;
  }
  if(v64)
    return (long long int)((valh | vall)&0xffffffffffffffffULL);
  else
    return (long long int)((valh | vall)&0x00000000ffffffffULL);
}

static long long int LSLReg(int m, int siz, int sft)
{
 long long int val;
 val = get_reg(m);
 switch(siz){
 case 0:
   break;
 case 1:
   if(sft)
     val = val << 1;
   break;
 case 2:
   if(sft)
     val = val << 2;
   break;
 case 3:
   if(sft)
     val = val << 3;
   break;
 }
 return val;
}

static long long int ExtendReg(int m, int ty, int sft)
{
  long long int val;
  int us;
  int len;
  val = get_reg(m);
  val = val << sft;
  switch(ty){
  case UXTB:
    us  = 0;
    len = 8;
    val &= 0xffLL;
    break;
  case UXTH:
    us  = 0;
    len = 16;
    val &= 0xffffLL;
    break;
  case UXTW:
    us  = 0;
    len = 32;
    val &= 0xffffffffLL;
    break;
  case UXTX:
    us  = 0;
    len = 64;
    val &= 0xffffffffffffffffLL;
    break;
  case SXTB:
    us  = 1;
    len = 8;
    val = (val & (1LL<<7))? 0xffffffffffffff00LL|val : 0xffLL&val;
    break;
  case SXTH:
    us  = 1;
    len = 16;
    val = (val & (1LL<<15))? 0xffffffffffff0000LL|val : 0xffffLL&val;
    break;
  case SXTW:
    us  = 1;
    len = 32;
    val = (val & (1LL<<31))? 0xffffffff00000000LL|val : 0xffffffffLL&val;
    break;
  case SXTX:
    us  = 1;
    len = 64;
    val &= 0xffffffffffffffffLL;
    break;
  }
  return val;
}

static void dec_undef(void)
{
  rs1 = NONE;
  rs2 = NONE;
  rs3 = NONE;
  rd  = NONE;
  OP  = NOP;
  OP2 = UDF;
}

static void dec_pc_rel(void)
{
  rd  = (inst_reg)&0x1f;
  rs1 = PC_VAL;
  rs2 = IMMD;
  immd = (long long int)((((inst_reg>>5)&0xfffffLL)<<2)|((inst_reg>>29)&3));
  immd = (immd & (1<<21))? 0xffffffffffc00000LL | immd : 0x3fffffLL & immd;
  v64 = 1;
  OP  = EADD;
  OP2 = NOP;
  if((inst_reg>>31)&1){/* ADRP */
    immd = (immd << 12);
    rs1 = PCP_VAL;
  }
  else{/* ADR */
  }
}

static void dec_add_sub_i(void)
{
  v64 = (inst_reg>>31)&1;
  rd  = inst_reg & 0x1f;
  rs1 = (inst_reg>>5)&0x1f;
  rs2 = IMMD;
  immd =(long long)((inst_reg>>10)&0xfff);
  if((inst_reg>>22)&1)
    immd = immd<<12;
  if(rs1==31)
    rs1 = SP_VAL;
  if((inst_reg>>30)&1)
    OP  = SUB;
  else
    OP  = ADD;
  if((inst_reg>>29)&1){
    OP2 = FLAG;
  }
  else {
    OP2 = NOP;
    if(rd==31)
      rd = SP_VAL;
  }
}

static long long int DecodeBitMasks(int N, int imms, int immr)
{
  unsigned long long int telem;
  unsigned long long int welem;
  unsigned long long int ror;
  int len;
  int levels;
  int S;
  int R;
  int diff;
  int esize;
  int d;
  int i;
  if(N==1){
    len = 6;
    esize = 64;
    levels = 0x3f;
  }
  else if((imms&0x20)==0){
    len = 5;
    esize = 32;
    levels = 0x1f;
  }
  else if((imms&0x10)==0){
    len = 4;
    esize = 16;
    levels = 0x0f;
  }
  else if((imms&0x8)==0){
    len = 3;
    esize = 8;
    levels = 0x7;
  }
  else if((imms&0x4)==0){
    len = 2;
    esize = 4;
    levels = 0x3;
  }
  else if((imms&0x2)==0){
    len = 1;
    esize = 2;
    levels = 0x1;
  }
  else if((imms&0x1)==0){
    len = 0;
    esize = 1;
    levels = 0x0;
  }
  S = imms & levels;
  R = immr & levels;
  diff = (S-R);
  d = diff & levels;
  if(S==0x3f){
    welem = 0x7fffffffffffffffULL;
  }
  else if(S==0x3e){
    welem = 0x3fffffffffffffffULL;
  }
  else{
    welem = (1ULL<<(S+1))-1ULL;
  }
  if(d==0x3f){
    telem = 0x7fffffffffffffffULL;
  }
  else if(d==0x3e){
    telem = 0x3fffffffffffffffULL;
  }
  else{
    telem = (1ULL<<(d+1))-1ULL;
  }
  ror   = ((welem<<(esize-R))|(welem>>R));
  switch(len){
  case 6:
    ror &= 0xffffffffffffffffULL;
    break;
  case 5:
    ror &= 0x00000000ffffffffULL;
    break;
  case 4:
    ror &= 0x000000000000ffffULL;
    break;
  case 3:
    ror &= 0x00000000000000ffULL;
    break;
  case 2:
    ror &= 0x000000000000000fULL;
    break;
  case 1:
    ror &= 0x0000000000000003ULL;
    break;
  default:
    ror &= 0x0000000000000001ULL;
    break;
  }
  wmask = 0ULL;
  tmask = 0ULL;
  for(i=0;i<(64/esize);i++){
    wmask = (wmask<<esize)|ror;
    tmask = (tmask<<esize)|telem;
  }
  return (long long int)wmask;
}

static void dec_logic_i(void)
{
  v64 = (inst_reg>>31)&1;
  rd  = inst_reg & 0x1f;
  rs1 = (inst_reg>>5)&0x1f;
  rs2 = IMMD;
  immd =DecodeBitMasks((inst_reg>>22)&1,(inst_reg>>10)&0x3f,(inst_reg>>16)&0x3f);
  switch((inst_reg>>29)&3){
  case 0:
    OP  = AND;
    OP2 = NOP;
    if(rd==31)
      rd = SP_VAL;
    break;
  case 1:
    OP  = OR;
    OP2 = NOP;
    if(rd==31)
      rd = SP_VAL;
    break;
  case 2:
    OP  = XOR;
    OP2 = NOP;
    if(rd==31)
      rd = SP_VAL;
    break;
  case 3:
    OP  = AND;
    OP2 = FLAG;
    break;
  }
}

static void dec_movw(void)
{
  int pos;
  v64 = (inst_reg>>31)&1;
  pos = (inst_reg>>21)&3;
  rd  = inst_reg & 0x1f;
  rs1 = NONE;
  rs2 = IMMD;
  immd =(long long int)((inst_reg>>5)&0xffff);
  op = (inst_reg>>29)&3;
  pos = pos<<4;
  switch(op){
  case 0:/* MOVN */
    immd = immd<<pos;
    immd = (~immd)&0xffffffffffffffffLL;
    OP  = MOV;
    OP2 = NOP;
    break;
  case 2:/* MOVZ */
    immd = immd<<pos;
    OP  = MOV;
    OP2 = NOP;
    break;
  case 3:/* MOVK */
    immd = immd<<pos;
    imask=(long long int)(0xffffLL<<pos);
    OP  = MOVK;
    OP2 = NOP;
    break;
  default:
    dec_undef();
    break;
  }
}

static void dec_bitf(void)
{
  v64 = (inst_reg>>31)&1;
  rd  = inst_reg & 0x1f;
  rs1 = (inst_reg>>5)&0x1f;
  rs2 = NONE;
  d_N = (inst_reg>>22)&1;
  d_S = (inst_reg>>10)&0x3f;
  d_R = (inst_reg>>16)&0x3f;
  immd =DecodeBitMasks(d_N,d_S,d_R);
  op  = (inst_reg>>29)&3;
  switch(op){
  case 0:/* SBFM */
    OP  = SBFM;
    OP2 = NOP;
    break;
  case 1:/* BFM */
    OP  = BFM;
    OP2 = NOP;
    break;
  case 2:/* UBFM */
    OP  = UBFM;
    OP2 = NOP;
    break;
  default:
    dec_undef();
    break;
  }
}

static void dec_extr(void)
{
  v64 = (inst_reg>>31)&1;
  rd  = inst_reg & 0x1f;
  rs1 = (inst_reg>>5)&0x1f;
  rs2 = (inst_reg>>16)&0x1f;
  d_N = (inst_reg>>22)&1;
  d_S = (inst_reg>>10)&0x3f;
  op  = (inst_reg>>29)&0x3;
  op2 = (inst_reg>>21)&1;
  if((op==0)&&(op2==0)){/* EXTR */
    OP  = EXTR;
    OP2 = NOP;
  }
  else{
    dec_undef();
  }
}

static void dec_data_proc_i(void)
{
  op = (inst_reg>>23)&0x7;
  switch(op){
  case 0x0:
  case 0x1:
    dec_pc_rel();
    break;
  case 0x2:
  case 0x3:
    dec_add_sub_i();
    break;
  case 0x4:
    dec_logic_i();
    break;
  case 0x5:
    dec_movw();
    break;
  case 0x6:
    dec_bitf();
    break;
  case 0x7:
    dec_extr();
    break;
  }
}

static void dec_data_proc_2src(void)
{
  int S;
  v64 = (inst_reg>>31)&1;
  rs1 = (inst_reg>> 5)&0x1f;
  rs2 = (inst_reg>>16)&0x1f;
  rd  = (inst_reg>> 0)&0x1f;
  op  = (inst_reg>>10)&0x3f;
  S   = (inst_reg>>29)&1;
  if((S==0)&&(op==0x02)){/* UDIV */
    OP  = UDIV;
    OP2 = NOP;
    return;
  }
  if((S==0)&&(op==0x03)){/* SDIV */
    OP  = SDIV;
    OP2 = NOP;
    return;
  }
  if((S==0)&&(op==0x08)){/* LSLV */
    OP  = SLL;
    OP2 = NOP;
    return;
  }
  if((S==0)&&(op==0x09)){/* LSRV */
    OP  = SRL;
    OP2 = NOP;
    return;
  }
  if((S==0)&&(op==0x09)){/* ASRV */
    OP  = SRA;
    OP2 = NOP;
    return;
  }
  if((S==0)&&(op==0x0b)){/* RORV */
    OP  = ROR;
    OP2 = NOP;
    return;
  }
  dec_undef();
}

static void dec_data_proc_1src(void)
{
  int S;
  v64 = (inst_reg>>31)&1;
  rs1 = (inst_reg>> 5)&0x1f;
  rd  = (inst_reg>> 0)&0x1f;
  op  = (inst_reg>>10)&0x3f;
  op2 = (inst_reg>>16)&0x1f;
  S   = (inst_reg>>29)&1;
  if((S==0)&&(op2==0)&&(op==0)){/* RBIT */
    OP  = RBIT;
    OP2 = NOP;
    return;
  }
  if((S==0)&&(op2==0)&&(op==1)){/* REV16 */
    OP  = REV16;
    OP2 = NOP;
    return;
  }
  if((v64==0)&&(S==0)&&(op2==0)&&(op==2)){/* REV */
    OP  = REV;
    OP2 = NOP;
    return;
  }
  if((v64==1)&&(S==0)&&(op2==0)&&(op==2)){/* REV32 */
    OP  = REV32;
    OP2 = NOP;
    return;
  }
  if((v64==1)&&(S==0)&&(op2==0)&&(op==3)){/* REV */
    OP  = REV;
    OP2 = NOP;
    return;
  }
  if((S==0)&&(op2==0)&&(op==4)){/* CLZ  */
    OP  = CLZ;
    OP2 = NOP;
    return;
  }
  if((S==0)&&(op2==0)&&(op==5)){/* CLS  */
    OP  = CLS;
    OP2 = NOP;
    return;
  }
  dec_undef();
}


static void dec_logical_shift(void)
{
  int N;
  int sa;
  int st;
  v64 = (inst_reg>>31)&1;
  op  = (inst_reg>>29)&3;
  st  = (inst_reg>>22)&3;
  N   = (inst_reg>>21)&1;
  rs1 = (inst_reg>> 5)&0x1f;
  rs2 = (inst_reg>>16)&0x1f;
  rd  = (inst_reg>> 0)&0x1f;
  sa  = (inst_reg>>10)&0x3f;
  immd = ShiftReg(rs2, st, sa);
  rs2 = IMMD;
  switch((op<<1)|N){
  case 0:
    OP  = AND;
    OP2 = NOP;
    break;
  case 1:
    OP  = BIC;
    OP2 = NOP;
    break;
  case 2:
    OP  = OR;
    OP2 = NOP;
    break;
  case 3:
    OP  = ORN;
    OP2 = NOP;
    break;
  case 4:
    OP  = XOR;
    OP2 = NOP;
    break;
  case 5:
    OP  = XORN;
    OP2 = NOP;
    break;
  case 6:
    OP  = AND;
    OP2 = FLAG;
    break;
  case 7:
    OP  = BIC;
    OP2 = FLAG;
    break;
  }
}

static void dec_addsub_shift(void)
{
  int S;
  int st;
  int sa;
  v64 = (inst_reg>>31)&1;
  op  = (inst_reg>>30)&1;
  S   = (inst_reg>>29)&1;
  st  = (inst_reg>>22)&3;
  rs1 = (inst_reg>> 5)&0x1f;
  rs2 = (inst_reg>>16)&0x1f;
  rd  = (inst_reg>> 0)&0x1f;
  sa  = (inst_reg>>10)&0x3f;
  immd = ShiftReg(rs2, st, sa);
  rs2 = IMMD;
  if(op)
    OP = SUB;
  else
    OP = ADD;
  if(S)
    OP2 = FLAG;
  else
    OP2 = NOP;
}

static void dec_addsub_extend(void)
{
  int S;
  int ty;
  int opt;
  int imm3;
  v64 = (inst_reg>>31)&1;
  op  = (inst_reg>>30)&1;
  S   = (inst_reg>>29)&1;
  ty  = (inst_reg>>22)&3;
  rs1 = (inst_reg>> 5)&0x1f;
  rs2 = (inst_reg>>16)&0x1f;
  rd  = (inst_reg>> 0)&0x1f;
  imm3 = (inst_reg>>10)&0x7;
  opt = (inst_reg>>13)&0x7;
  if(rs1==31)
    rs1 = SP_VAL;
  immd = ExtendReg(rs2, DecodeRegExtend(opt), imm3);
  rs2 = IMMD;
  if(op){
    OP = SUB;
    if(rd==31)
      rd = (S)? 31 : SP_VAL;
  }
  else {
    OP = ADD;
    if(rd==31)
      rd = (S)? 31 : SP_VAL;
  }
  if(S)
    OP2 = FLAG;
  else
    OP2 = NOP;
}

static void dec_addsub_carry(void)
{
  int S;
  int opt;
  v64 = (inst_reg>>31)&1;
  op  = (inst_reg>>30)&1;
  S   = (inst_reg>>29)&1;
  rs1 = (inst_reg>> 5)&0x1f;
  rs2 = (inst_reg>>16)&0x1f;
  rd  = (inst_reg>> 0)&0x1f;
  op2 = (inst_reg>>10)&0x3f;
  if(op2==0){
    if(op)
      OP = SBC;
    else
      OP = ADC;
    if(S)
      OP2 = FLAG;
    else
      OP2 = NOP;
  }
  else{
    dec_undef();
  }
}


static void dec_cond_comp_r(void)
{
  int S;
  int opt;
  v64 = (inst_reg>>31)&1;
  op  = (inst_reg>>30)&1;
  op2 = (inst_reg>>10)&1;
  op3 = (inst_reg>> 4)&1;
  S   = (inst_reg>>29)&1;
  rs1 = (inst_reg>> 5)&0x1f;
  rs2 = (inst_reg>>16)&0x1f;
  nzcv = (inst_reg>> 0)&0xf;
  cond = (inst_reg>>12)&0xf;
  rd  = NONE;
  if((S==0)||(op2==1)||(op3==1)){
    dec_undef();
    return;
  }
  if(op)
    OP = CCMP;
  else
    OP = CCMN;
  OP2 = FLAG;
}


static void dec_cond_comp_i(void)
{
  int S;
  int opt;
  v64 = (inst_reg>>31)&1;
  op  = (inst_reg>>30)&1;
  op2 = (inst_reg>>10)&1;
  op3 = (inst_reg>> 4)&1;
  S   = (inst_reg>>29)&1;
  rs1 = (inst_reg>> 5)&0x1f;
  rs2 = IMMD;
  nzcv = (inst_reg>> 0)&0xf;
  cond = (inst_reg>>12)&0xf;
  rd  = NONE;
  immd = (long long int)((inst_reg>>16)&0x1f);
  if((S==0)||(op2==1)||(op3==1)){
    dec_undef();
    return;
  }
  if(op)
    OP = CCMP;
  else
    OP = CCMN;
  OP2 = FLAG;
}


static void dec_cond_select(void)
{
  int S;
  int opt;
  v64 = (inst_reg>>31)&1;
  op  = (inst_reg>>30)&1;
  op2 = (inst_reg>>10)&3;
  S   = (inst_reg>>29)&1;
  rs1 = (inst_reg>> 5)&0x1f;
  rs2 = (inst_reg>>16)&0x1f;
  rd  = (inst_reg>> 0)&0x1f;
  cond = (inst_reg>>12)&0xf;
  if((S==1)||(op2==2)||(op2==3)){
    dec_undef();
    return;
  }
  switch((op<<1)|op2){
  case 0:
    OP = CSEL;
    break;
  case 1:
    OP = CSINC;
    break;
  case 2:
    OP = CSINV;
    break;
  case 3:
    OP = CSNEG;
    break;
  }
  OP2 = NOP;
}

static void dec_data_proc_3src(void)
{
  v64 = (inst_reg>>31)&1;
  op  = (inst_reg>>29)&3;
  op2 = (inst_reg>>21)&7;
  op3 = (inst_reg>>15)&1;
  rs1 = (inst_reg>> 5)&0x1f;
  rs2 = (inst_reg>>16)&0x1f;
  rs3 = (inst_reg>>10)&0x1f;
  rd  = (inst_reg>> 0)&0x1f;
  if((op==0)&&(op2==0)&&(op3==0)){
    OP  = MADD;
    OP2 = NOP;
    return;
  }
  if((op==0)&&(op2==0)&&(op3==1)){
    OP  = MSUB;
    OP2 = NOP;
    return;
  }
  if((v64==1)&&(op==0)&&(op2==1)&&(op3==0)){
    OP  = SMADDL;
    OP2 = NOP;
    return;
  }
  if((v64==1)&&(op==0)&&(op2==1)&&(op3==1)){
    OP  = SMSUBL;
    OP2 = NOP;
    return;
  }
  if((v64==1)&&(op==0)&&(op2==2)&&(op3==0)){
    OP  = SMULH;
    OP2 = NOP;
    return;
  }
  if((v64==1)&&(op==0)&&(op2==5)&&(op3==0)){
    OP  = UMADDL;
    OP2 = NOP;
    return;
  }
  if((v64==1)&&(op==0)&&(op2==5)&&(op3==1)){
    OP  = UMSUBL;
    OP2 = NOP;
    return;
  }
  if((v64==1)&&(op==0)&&(op2==6)&&(op3==0)){
    OP  = UMULH;
    OP2 = NOP;
    return;
  }
  dec_undef();
}

static void dec_data_proc_r(void)
{
  op  = (inst_reg>>30)&1;
  op1 = (inst_reg>>28)&1;
  op2 = (inst_reg>>21)&0xf;
  op3 = (inst_reg>>11)&1;
  if((op==0)&&(op1==1)&&(op2==6)){
    dec_data_proc_2src();
    return;
  }
  if((op==1)&&(op1==1)&&(op2==6)){
    dec_data_proc_1src();
    return;
  }
  if((op1==0)&&((op2&8)==0)){
    dec_logical_shift();
    return;
  }
  if((op1==0)&&((op2&9)==8)){
    dec_addsub_shift();
    return;
  }
  if((op1==0)&&((op2&9)==9)){
    dec_addsub_extend();
    return;
  }
  if((op1==1)&&(op2==0)){
    dec_addsub_carry();
    return;
  }
  if((op1==1)&&(op2==2)&&(op3==0)){
    dec_cond_comp_r();
    return;
  }
  if((op1==1)&&(op2==2)&&(op3==1)){
    dec_cond_comp_i();
    return;
  }
  if((op1==1)&&(op2==4)){
    dec_cond_select();
    return;
  }
  if((op1==1)&&((op2&8)==8)){
    dec_data_proc_3src();
    return;
  }
  dec_undef();
}

static void dec_cond_bra(ovid)
{
  v64 = 1;
  op  = (inst_reg>>24)&1;
  op2 = (inst_reg>>4)&1;
  if((op==0)&&(op2==0)){
    cond = inst_reg & 0xf;
    rd = NONE;
    rs1 = PC_VAL;
    rs2 = IMMD;
    immd = (long long int)((inst_reg>>5)&0x7ffff);
    immd = (immd & (1LL<<18))? 0xfffffffffff80000LL|immd : 0x7ffffLL&immd;
    immd = immd<<2;
    OP  = EADD;
    OP2 = JCC;
  }
  else{
    dec_undef();
  }
}

static void dec_except(void)
{
  dec_undef();/* not supported */
}

static void dec_system(void)
{
  if(inst_reg==0xd503201f){
    OP = NOP;
    OP2 = NOP;
    rd  = NONE;
    return;
  }
  dec_undef();/* not supported */
}

static void dec_uncond_bra_r(void)
{
  int Z, M;
  op =  (inst_reg>>21)&0xf;
  op2 = (inst_reg>>16)&0x1f;
  op3 = (inst_reg>>10)&0x3f;
  op4 = (inst_reg>> 0)&0x1f;
  Z   = (inst_reg>>24)&1;
  M   = (inst_reg>>10)&1;
  rs1 = (inst_reg>> 5)&0x1f;
  rs2 = (inst_reg>> 0)&0x1f;
  rd  = NONE;
  //  use_key_a = (M==0);
  //  source_is_sp = ((Z==1)&&(rs2==31));
  if((op==0)&&(op2==31)&&(op3==0)&&(op4==0)){/* BR */
    OP  = NOP;
    OP2 = JPR;
  }
  else if((op==0)&&(op2==31)&&(op3==2)&&(op4==31)){/* BRA */
    dec_undef();/* not supported */    
  }
  else if((op==0)&&(op2==31)&&(op3==3)&&(op4==31)){/* BRB */
    dec_undef();/* not supported */
  }
  else if((op==1)&&(op2==31)&&(op3==0)&&(op4==0)){/* BLR */
    rs2 = NXT_PC;
    rd  = 30;
    OP  = MOV;
    OP2 = JPR;
  }
  else if((op==1)&&(op2==31)&&(op3==2)&&(op4==31)){/* BLRA */
    dec_undef();/* not supported */
  }
  else if((op==1)&&(op2==31)&&(op3==3)&&(op4==31)){/* BLRB */
    dec_undef();/* not supported */
  }
  else if((op==2)&&(op2==31)&&(op3==0)&&(op4==0)){/* RET */
    OP  = NOP;
    OP2 = JPR;
  }
  else if((op==2)&&(op2==31)&&(op3==2)&&(op4==31)&&(rs1==31)){/* RETA */
    dec_undef();/* not supported */
  }
  else if((op==2)&&(op2==31)&&(op3==3)&&(op4==31)&&(rs1==31)){/* RETB */
    dec_undef();/* not supported */
  }
  else{
    dec_undef();/* not supported */
  }
}

static void dec_uncond_bra_i(void)
{
  immd = (long long int)(inst_reg & 0x3ffffffLL);
  immd = (immd & (1LL<<25))? 0xfffffffffc000000LL|immd : 0x3ffffffLL&immd;
  immd = immd<<2;
  rs1  = PC_VAL;
  rs2  = IMMD;
  rd   = NONE;
  if((inst_reg>>31)&1){
    OP  = EADD;
    OP2 = JAL;
  }
  else{
    OP  = EADD;
    OP2 = JMP;
  }
}

static void dec_cmp_bra(void)
{
  v64 = (inst_reg>>31)&1;
  op  = (inst_reg>>24)&1;
  immd = (long long int)((inst_reg>>5)&0x7ffff);
  immd = (immd & (1LL<<18))? 0xfffffffffff80000LL|immd : 0x7ffffLL&immd;
  immd = immd<<2;
  rs1 = PC_VAL;
  rs2 = IMMD;
  rs3 = inst_reg & 0x1f;
  rd  = NONE;
  OP  = EADD;
  if(op==0)
    OP2 = CBZ;
  else
    OP2 = CBNZ;
}

static void dec_tst_bra(void)
{
  v64 = (inst_reg>>31)&1;
  op  = (inst_reg>>24)&1;
  bit_pos = ((inst_reg>>19)&0x1f)|(v64<<5);
  immd = (long long int)((inst_reg>>5)&0x3fff);
  immd = (immd & (1LL<<13))? 0xffffffffffffc000LL|immd : 0x3fffLL&immd;
  immd = immd<<2;
  rs1 = PC_VAL;
  rs2 = IMMD;
  rs3 = inst_reg & 0x1f;
  rd  = NONE;
  OP  = EADD;
  if(op==0)
    OP2 = TBZ;
  else
    OP2 = TBNZ;
}

static void dec_branch_sys(void)
{
  op  = (inst_reg>>29)&0x7;
  op1 = (inst_reg>>22)&0xf;
  if((op==2)&&((op1&8)==0)){
    dec_cond_bra();
  }
  else if((op==2)&&((op1&8)==8)){
    dec_undef();
  }
  else if((op==6)&&((op1&12)==0)){
    dec_except();
  }
  else if((op==6)&&(op1==4)){
    dec_system();
  }
  else if((op==6)&&(op1==5)){
    dec_undef();
  }
  else if((op==6)&&((op1&14)==6)){
    dec_undef();
  }
  else if((op==6)&&((op1&8)==8)){
    dec_uncond_bra_r();
  }
  else if((op&3)==0){
    dec_uncond_bra_i();
  }
  else if(((op&3)==1)&&((op2&8)==0)){
    dec_cmp_bra();
  }
  else if(((op&3)==1)&&((op2&8)==8)){
    dec_tst_bra();
  }
  else{
    dec_undef();
  }
}

static void dec_ldst(void)
{
  int V,L,S;
  op  = (inst_reg>>31)&0x1;
  op1 = (inst_reg>>28)&3;
  op2 = (inst_reg>>26)&1;
  op3 = (inst_reg>>23)&3;
  op4 = (inst_reg>>16)&0x3f;
  op5 = (inst_reg>>10)&3;
  V   = (inst_reg>>26)&1;
  L   = (inst_reg>>22)&1;
  post_index = 0;
  mwback = 0;
  if((op1==0)&&(op2==0)&&((op3&2)==0)){/* LDREX/STREX */
    dec_undef();/* not supported */
    return;
  }
  if((op1==1)&&((op3&2)==0)){/* LDRPC */
    op  = (inst_reg>>30)&3;
    rd  = inst_reg & 0x1f;
    immd = (long long int)((inst_reg>>5)&0x7ffff);
    immd = (immd & (1LL<<18))? 0xfffffffffff80000LL|immd : 0x7ffffLL&immd;
    immd = immd << 2;
    if(V){
      dec_undef();/* not supported */
    }
    else{
      rs1 = PC_VAL;
      rs2 = IMMD;
      OP  = EADD;
      OP2 = LDPC;
      switch(op){
      case 0:
	msize = WORD;
	msex  = 0;
	break;
      case 1:
	msize = DWORD;
	msex  = 0;
	break;
      case 2:
	msize = WORD;
	msex  = 1;
	break;
      case 3:/* prefetch */
	break;
      }
    }
    return;
  }
  if((op1==2)&&(op3==0)){/* LDNP/STNP(offset) */
    op  = (inst_reg>>30)&3;
    rd  = (inst_reg>> 0)&0x1f;
    rd2 = (inst_reg>>10)&0x1f;
    rs1 = (inst_reg>> 5)&0x1f;
    if(rs1==31)
      rs1 = SP_VAL;
    rs2 = IMMD;
    OP  = EADD;
    if(L){
      OP2 = LDNP;
    }
    else{
      OP2 = STNP;
      rs3 = rd;
      rs4 = rd2;
      rd  = NONE;
      rd2 = NONE;
    }
    if(V){
      dec_undef();/* not supported */
    }
    else{
      immd = (long long int)((inst_reg>>15)&0x7f);
      immd = (immd & (1LL<<6))? 0xffffffffffffff80LL|immd : 0x7fLL&immd;
      switch(op){
      case 0:
	immd = immd << 2;
	msize = WORD;
	msex  = 0;
	break;
      case 2:
	immd = immd << 3;
	msize = DWORD;
	msex  = 0;
	break;
      case 1:
      case 3:
	dec_undef();
	break;
      }
    }
    return;
  }
  if((op1==2)&&(op3==1)){/* LDRP/STRP(post-indexed) */
    op  = (inst_reg>>30)&3;
    rd  = (inst_reg>> 0)&0x1f;
    rd2 = (inst_reg>>10)&0x1f;
    rs1 = (inst_reg>> 5)&0x1f;
    if(rs1==31)
      rs1 = SP_VAL;
    rs2 = IMMD;
    OP  = EADD;
    if(L){
      OP2 = LDP;
    }
    else{
      OP2 = STP;
      rs3 = rd;
      rs4 = rd2;
      rd  = NONE;
      rd2 = NONE;
    }
    if(V){
      dec_undef();/* not supported */
    }
    else{
      immd = (long long int)((inst_reg>>15)&0x7f);
      immd = (immd & (1LL<<6))? 0xffffffffffffff80LL|immd : 0x7fLL&immd;
      post_index = 1;
      mwback = 1;
      switch(op){
      case 0:
	immd = immd << 2;
	msize = WORD;
	msex  = 0;
	break;
      case 1:
	immd = immd << 2;
	if(L){
	  msize = WORD;
	  msex  = 1;
	}
	else{
	  dec_undef();
	}
	break;
      case 2:
	immd = immd << 3;
	msize = DWORD;
	msex  = 0;
	break;
      case 3:
	dec_undef();
	break;
      }
    }
    return;
  }
  if((op1==2)&&(op3==2)){/* LDRP/STRP(offset) */
    op  = (inst_reg>>30)&3;
    rd  = (inst_reg>> 0)&0x1f;
    rd2 = (inst_reg>>10)&0x1f;
    rs1 = (inst_reg>> 5)&0x1f;
    if(rs1==31)
      rs1 = SP_VAL;
    rs2 = IMMD;
    OP  = EADD;
    if(L){
      OP2 = LDP;
    }
    else{
      OP2 = STP;
      rs3 = rd;
      rs4 = rd2;
      rd  = NONE;
      rd2 = NONE;
    }
    if(V){
      dec_undef();/* not supported */
    }
    else{
      immd = (long long int)((inst_reg>>15)&0x7f);
      immd = (immd & (1LL<<6))? 0xffffffffffffff80LL|immd : 0x7fLL&immd;
      post_index = 0;
      mwback = 0;
      switch(op){
      case 0:
	immd = immd << 2;
	msize = WORD;
	msex  = 0;
	break;
      case 1:
	immd = immd << 2;
	if(L){
	  msize = WORD;
	  msex  = 1;
	}
	else{
	  dec_undef();
	}
	break;
      case 2:
	immd = immd << 3;
	msize = DWORD;
	msex  = 0;
	break;
      case 3:
	dec_undef();
	break;
      }
    }
    return;
  }
  if((op1==2)&&(op3==3)){/* LDRP/STRP(pre-indexed) */
    op  = (inst_reg>>30)&3;
    rd  = (inst_reg>> 0)&0x1f;
    rd2 = (inst_reg>>10)&0x1f;
    rs1 = (inst_reg>> 5)&0x1f;
    if(rs1==31)
      rs1 = SP_VAL;
    rs2 = IMMD;
    OP  = EADD;
    if(L){
      OP2 = LDP;
    }
    else{
      OP2 = STP;
      rs3 = rd;
      rs4 = rd2;
      rd  = NONE;
      rd2 = NONE;
    }
    if(V){
      dec_undef();/* not supported */
    }
    else{
      immd = (long long int)((inst_reg>>15)&0x7f);
      immd = (immd & (1LL<<6))? 0xffffffffffffff80LL|immd : 0x7fLL&immd;
      post_index = 0;
      mwback = 1;
      switch(op){
      case 0:
	immd = immd << 2;
	msize = WORD;
	msex  = 0;
	break;
      case 1:
	immd = immd << 2;
	if(L){
	  msize = WORD;
	  msex  = 1;
	}
	else{
	  dec_undef();
	}
	break;
      case 2:
	immd = immd << 3;
	msize = DWORD;
	msex  = 0;
	break;
      case 3:
	dec_undef();
	break;
      }
    }
    return;
  }
  if((op1==3)&&((op3&2)==0)&&((op4&0x20)==0)&&(op5==0)){/* LDR/STR(imm) */
    op  = (inst_reg>>22)&3;
    rd  = (inst_reg>> 0)&0x1f;
    rs1 = (inst_reg>> 5)&0x1f;
    rs2 = IMMD;
    if(rs1==31)
      rs1 = SP_VAL;
    immd = (long long int)((inst_reg>>12)&0x1ff);
    immd = (immd & (1LL<<8))? 0xfffffffffffffe00LL|immd : 0x1ffLL&immd;
    OP = EADD;
    post_index = 0;
    mwback = 0;
    switch((inst_reg>>30)&3){
    case 0:
      msize = BYTE;
      break;
    case 1:
      msize = HWORD;
      break;
    case 2:
      msize = WORD;
      break;
    case 3:
      msize = DWORD;
      break;
    }
    if(V){
      dec_undef();/* not supported */
    }
    else{
      switch(op){
      case 0:
	OP2 = STUR;
	rs3 = rd;
	rd  = NONE;
	msex = 0;
	break;
      case 1:
	OP2 = LDUR;
	msex = 0;
	break;
      case 2:
	if(msize==DWORD){/* PRFM */
	  rd  = NONE;
	  OP2 = NOP;
	}
	else{
	  OP2 = LDUR;
	  msex = 1;
	}
	break;
      case 3:
	if((msize==BYTE)||(msize==HWORD)){
	  OP2 = LDUR;
	  msex = 1;
	}
	else{
	  dec_undef();
	}
	break;
      }
    }
    return;
  }
  if((op1==3)&&((op3&2)==0)&&((op4&0x20)==0)&&(op5==1)){/* LDR/STR(imm,post-indexed) */
    op  = (inst_reg>>22)&3;
    rd  = (inst_reg>> 0)&0x1f;
    rs1 = (inst_reg>> 5)&0x1f;
    rs2 = IMMD;
    if(rs1==31)
      rs1 = SP_VAL;
    immd = (long long int)((inst_reg>>12)&0x1ff);
    immd = (immd & (1LL<<8))? 0xfffffffffffffe00LL|immd : 0x1ffLL&immd;
    OP = EADD;
    post_index = 1;
    mwback = 1;
    switch((inst_reg>>30)&3){
    case 0:
      msize = BYTE;
      break;
    case 1:
      immd = immd << 1;
      msize = HWORD;
      break;
    case 2:
      immd = immd << 2;
      msize = WORD;
      break;
    case 3:
      immd = immd << 3;
      msize = DWORD;
      break;
    }
    if(V){
      dec_undef();/* not supported */
    }
    else{
      switch(op){
      case 0:
	OP2 = ST;
	rs3 = rd;
	rd  = NONE;
	msex = 0;
	break;
      case 1:
	OP2 = LD;
	msex = 0;
	break;
      case 2:
	if(msize==DWORD){
	  dec_undef();
	}
	else{
	  OP2 = LD;
	  msex = 1;
	}
	break;
      case 3:
	if((msize==BYTE)||(msize==HWORD)){
	  OP2 = LD;
	  msex = 1;
	}
	else{
	  dec_undef();
	}
	break;
      }
    }
    return;
  }
  if((op1==3)&&((op3&2)==0)&&((op4&0x20)==0)&&(op5==2)){/* LDR/STR(unpriv) */
    dec_undef();/* not supported */
    return;
  }
  if((op1==3)&&((op3&2)==0)&&((op4&0x20)==0)&&(op5==3)){/* LDR/STR(imm, pre-indexed) */
    op  = (inst_reg>>22)&3;
    rd  = (inst_reg>> 0)&0x1f;
    rs1 = (inst_reg>> 5)&0x1f;
    rs2 = IMMD;
    if(rs1==31)
      rs1 = SP_VAL;
    immd = (long long int)((inst_reg>>12)&0x1ff);
    immd = (immd & (1LL<<8))? 0xfffffffffffffe00LL|immd : 0x1ffLL&immd;
    OP = EADD;
    post_index = 0;
    mwback = 1;
    switch((inst_reg>>30)&3){
    case 0:
      msize = BYTE;
      break;
    case 1:
      immd = immd << 1;
      msize = HWORD;
      break;
    case 2:
      immd = immd << 2;
      msize = WORD;
      break;
    case 3:
      immd = immd << 3;
      msize = DWORD;
      break;
    }
    if(V){
      dec_undef();/* not supported */
    }
    else{
      switch(op){
      case 0:
	OP2 = ST;
	rs3 = rd;
	rd  = NONE;
	msex = 0;
	break;
      case 1:
	OP2 = LD;
	msex = 0;
	break;
      case 2:
	if(msize==DWORD){
	  dec_undef();
	}
	else{
	  OP2 = LD;
	  msex = 1;
	}
	break;
      case 3:
	if((msize==BYTE)||(msize==HWORD)){
	  OP2 = LD;
	  msex = 1;
	}
	else{
	  dec_undef();
	}
	break;
      }
    }
    return;
  }
  if((op1==3)&&((op3&2)==0)&&((op4&0x20)==0x20)&&(op5==0)){/* Atomic */
    dec_undef();/* not supported */
    return;
  }
  if((op1==3)&&((op3&2)==0)&&((op4&0x20)==0x20)&&(op5==2)){/* LDR/STR(reg offset) */
    int option;
    int ext_typ;
    op  = (inst_reg>>22)&3;
    rd  = (inst_reg>> 0)&0x1f;
    rs1 = (inst_reg>> 5)&0x1f;
    rs2 = (inst_reg>>16)&0x1f;
    S   = (inst_reg>>12)&1;
    option = (inst_reg>>13)&7;
    ext_typ = DecodeRegExtend(option);
    if(rs1==31)
      rs1 = SP_VAL;
    if(option == 3)
      immd = LSLReg(rs2, (inst_reg>>30)&3, S);
    else
      immd = ExtendReg(rs2, ext_typ, 0);
    rs2  = IMMD;
    OP = EADD;
    post_index = 0;
    mwback = 0;
    switch((inst_reg>>30)&3){
    case 0:
      msize = BYTE;
      break;
    case 1:
      if((option!=3)&&S){
	immd <<= 1;
      }
      msize = HWORD;
      break;
    case 2:
      if((option!=3)&&S){
	immd <<= 2;
      }
      msize = WORD;
      break;
    case 3:
      if((option!=3)&&S){
	immd <<= 3;
      }
      msize = DWORD;
      break;
    }
    if(V){
      dec_undef();/* not supported */
    }
    else{
      switch(op){
      case 0:
	OP2 = ST;
	rs3 = rd;
	rd  = NONE;
	msex = 0;
	break;
      case 1:
	OP2 = LD;
	msex = 0;
	break;
      case 2:
	if(msize==DWORD){
	  dec_undef();
	}
	else{
	  OP2 = LD;
	  msex = 1;
	}
	break;
      case 3:
	if((msize==BYTE)||(msize==HWORD)){
	  OP2 = LD;
	  msex = 1;
	}
	else{
	  dec_undef();
	}
	break;
      }
    }
    return;
  }
  if((op1==3)&&((op3&2)==0)&&((op4&0x20)==0x20)&&((op5&1)==1)){/* LDR/STR(pac) */
    dec_undef();/* not supported */
    return;
  }
  if((op1==3)&&((op3&2)==2)){/* LDR/STR(unsigned,imm) */
    op  = (inst_reg>>22)&3;
    rd  = (inst_reg>> 0)&0x1f;
    rs1 = (inst_reg>> 5)&0x1f;
    rs2 = IMMD;
    if(rs1==31)
      rs1 = SP_VAL;
    immd = (long long int)((inst_reg>>10)&0xfff);
    OP = EADD;
    switch((inst_reg>>30)&3){
    case 0:
      msize = BYTE;
      break;
    case 1:
      immd = immd << 1;
      msize = HWORD;
      break;
    case 2:
      immd = immd << 2;
      msize = WORD;
      break;
    case 3:
      immd = immd << 3;
      msize = DWORD;
      break;
    }
    if(V){
      dec_undef();/* not supported */
    }
    else{
      switch(op){
      case 0:
	OP2 = ST;
	rs3 = rd;
	rd  = NONE;
	msex = 0;
	break;
      case 1:
	OP2 = LD;
	msex = 0;
	break;
      case 2:
	if(msize==DWORD){
	  dec_undef();
	}
	else{
	  OP2 = LD;
	  msex = 1;
	}
	break;
      case 3:
	if((msize==BYTE)||(msize==HWORD)){
	  OP2 = LD;
	  msex = 1;
	}
	else{
	  dec_undef();
	}
	break;
      }
    }
    return;
  }
  dec_undef();/* not supported */
}

static void dec_fpu_simd(void)
{
  dec_undef();/* not supported */
}

void decode(void)
{
  opcode  = (inst_reg>>25)&0xf;
  v64 = 1;
  rd  = NONE;
  rd2 = NONE;
  rs1 = NONE;
  rs2 = NONE;
  rs3 = NONE;
  rs4 = NONE;
  switch(opcode){
  case 0x8:
  case 0x9:
    dec_data_proc_i();
    break;
  case 0xa:
  case 0xb:
    dec_branch_sys();
    break;
  case 0x4:
  case 0x6:
  case 0xc:
  case 0xe:
    dec_ldst();
    break;
  case 0x5:
  case 0xd:
    dec_data_proc_r();
    break;
  case 0x7:
  case 0xf:
    dec_fpu_simd();
    break;
  default:
    dec_undef();
    break;
  }
}
