#include "define.h"
int cond;
int immd;
int rs1;
int rs2;
int rs3;
int rd;
int OP;
int OP2;
int msize;
int msex;
int n_high;
int m_high;
int mlist;
int madd;
int mindex;
int mwback;
int amode;
int muser;
int mexret;
int word_hi;
int mincr;
int mrot;
int mlsb;
int mwidth;

int inst_reg;

static int opcode;
static int opcode2;
static int op;
static int op1;
static int op2;

extern int C_flag;

int ISFT32(int src, int sa, int* n, int* z, int* c, int* v);

static void dec_undef(void)
{
	rs1 = NONE;
	rs2 = NONE;
	rs3 = NONE;
	rd  = NONE;
	OP  = NOP;
	OP2 = UDF;
}

static void dec_uncond(void)
{
	op1 = (inst_reg>>20)&0xff;
	rs1 = (inst_reg>>16)&0xf;
	dec_undef(); /* not supported */
}

static int make_op(op)
{
  switch(op){
  case 0x0:
    return AND;
  case 0x1:
    return XOR;
  case 0x2:
    return SUB;
  case 0x3:
    return RSB;
  case 0x4:
    return ADD;
  case 0x5:
    return ADC;
  case 0x6:
    return SBC;
  case 0x7:
    return RSC;
  case 0x8:
    rd = NONE;
    return TST;
  case 0x9:
    rd = NONE;
    return TEQ;
  case 0xa:
    rd = NONE;
    return CMP;
  case 0xb:
    rd = NONE;
    return CMN;
  case 0xc:
    return OR;
  case 0xd:
    return MOV;
  case 0xe:
    return BIC;
  case 0xf:
    return MVN;
  }
}

static int make_immd12(int x)
{
	int dat,sa;
	int nf,zf,vf;
	dat = x & 0xff;
	sa  = (x>>8) & 0xf;
	sa  = sa<<1;
	OP  = ROR;
	return ISFT32(dat, sa, &nf, &zf, &C_flag, &vf);
}

static int make_ishift(int x)
{
	int dat, rm, st, sa;
	int nf,zf,vf;
	rm = x & 0xf;
	sa = (x>>7)&0x1f;
	st = (x>>5)&0x3;
	dat = get_reg(rm);
	if(st==0)
		OP = SLL;
	else if(st==1)
		OP = SRL;
	else if(st==2)
		OP = SRA;
	else if(sa==0)
		OP = RRX;
	else
		OP = ROR;
	return ISFT32(dat, sa, &nf, &zf, &C_flag, &vf);
}

static int make_rshift(int x)
{
	int dat, rm, rs, st, sa;
	int nf,zf,vf;
	rm = x & 0xf;
	rs = (x>>8)&0xf;
	st = (x>>5)&0x3;
	dat = get_reg(rm);
	sa  = get_reg(rs)&0x1f;
	if(st==0)
		OP = SLL;
	else if(st==1)
		OP = SRL;
	else if(st==2)
		OP = SRA;
	else
		OP = ROR;
	return ISFT32(dat, sa, &nf, &zf, &C_flag, &vf);
}

static void dec_data_proc_i(void)
{
  int tmp_op;
  tmp_op = (inst_reg>>21)&0xf;
  immd = make_immd12(inst_reg&0xfff);
  rs1  = (inst_reg>>16)&0xf;
  rs2  = IMMD;
  rd   = (inst_reg>>12)&0xf;
  OP   = make_op(tmp_op);
  if((inst_reg>>20)&0x1)
    OP2 = FLAG;
  else
    OP2 = ((tmp_op>=8)&&(tmp_op<=0xb))? FLAG : NOP;
}

static void dec_data_proc_r(void)
{
  int tmp_op;
  tmp_op = (inst_reg>>21)&0xf;
  immd = make_ishift(inst_reg&0xfff);
  rs1  = (inst_reg>>16)&0xf;
  rs2  = IMMD;
  rd   = (inst_reg>>12)&0xf;
  OP   = make_op(tmp_op);
  if((inst_reg>>20)&0x1)
    OP2 = FLAG;
  else
    OP2 = ((tmp_op>=8)&&(tmp_op<=0xb))? FLAG : NOP;
}

static void dec_data_proc_sr(void)
{
  int tmp_op;
  tmp_op = (inst_reg>>21)&0xf;
  immd = make_rshift(inst_reg&0xfff);
  rs1  = (inst_reg>>16)&0xf;
  rs2  = IMMD;
  rd   = (inst_reg>>12)&0xf;
  OP   = make_op(tmp_op);
  if((inst_reg>>20)&0x1)
    OP2 = FLAG;
  else
    OP2 = ((tmp_op>=8)&&(tmp_op<=0xb))? FLAG : NOP;
}

static void dec_misc(void)
{
	op  = (inst_reg>>21)&0x3;
	op1 = (inst_reg>>16)&0xf;
	op2 = (inst_reg>> 4)&0x7;
	switch(op2){
	case 0x0:/* notsupported MRS/MSR */
		rs1 = NONE;
		rs2 = NONE;
		rs3 = NONE;
		rd  = NONE;
		OP  = NOP; /* not supported */
		OP2 = NOP; /* not supported */
		break;
	case 0x1:
		if(op==1){/* BX */
			rs1 = inst_reg&0xf;
			rs2 = NONE;
			rd  = NONE;
			OP  = NOP;
			OP2 = JPR;
		}
		else if(op==3){/* CLZ */
			rs1 = inst_reg&0xf;
			rs2 = NONE;
			rd  = (inst_reg>>12)&0xf;
			OP  = CLZ;
			OP2 = NOP;
		}
		else{
			rs1 = NONE;
			rs2 = NONE;
			rs3 = NONE;
			rd  = NONE;
			OP  = NOP; /* not supported */
			OP2 = UDF; /* not supported */
		}
		break;
	case 0x2:/* not supported BXJ */
		rs1 = NONE;
		rs2 = NONE;
		rs3 = NONE;
		rd  = NONE;
		OP  = NOP; /* not supported */
		OP2 = NOP; /* not supported */
		break;
	case 0x3:/* BLX */
		rs1 = inst_reg&0xf;
		rs2 = NEXT_PC;
		rd  = 14;
		OP  = MOV;
		OP2 = JPR;
		break;
	case 0x5:/* SAT ADD/SUB */
		rs1 = NONE;
		rs2 = NONE;
		rs3 = NONE;
		rd  = NONE;
		OP  = NOP; /* not supported */
		OP2 = NOP; /* not supported */
		break;
	case 0x6:/* not supported ERET */
		rs1 = NONE;
		rs2 = NONE;
		rs3 = NONE;
		rd  = NONE;
		OP  = NOP; /* not supported */
		OP2 = NOP; /* not supported */
		break;
	case 0x7:/* BKPT/HVC/SMC */
		rs1 = NONE;
		rs2 = NONE;
		rs3 = NONE;
		rd  = NONE;
		OP  = NOP; /* not supported */
		OP2 = NOP; /* not supported */
		break;
	default:
		dec_undef();
		break;
	}
}

static void dec_mulh(void)
{
	op  = (inst_reg>>5)&0x1;
	op1 = (inst_reg>>21)&0x3;
	rd  = (inst_reg>>16)&0xf;
	rs1 = (inst_reg>> 0)&0xf;
	rs2 = (inst_reg>> 8)&0xf;
	rs3 = (inst_reg>>12)&0xf;
	n_high = (inst_reg>>5)&1;
	m_high = (inst_reg>>6)&0x1;
	switch(op1){
	case 0x0:
		OP  = SMLAxy;
		OP2 = NOP;
		break;	
	case 0x1:
		if(op==0)
			OP  = SMLAWxy;
		else
			OP  = SMULWxy;
		OP2 = NOP;
		break;	
	case 0x2:
		OP  = SMLALxy;
		OP2 = NOP;
		break;	
	case 0x3:
		OP  = SMULxy;
		OP2 = NOP;
		break;	
	}
}

static void dec_mul(void)
{
	op  = (inst_reg>>20)&0xf;
	rd  = (inst_reg>>16)&0xf;
	rs1 = (inst_reg>> 0)&0xf;
	rs2 = (inst_reg>> 8)&0xf;
	rs3 = (inst_reg>>12)&0xf;
	switch(op){
	case 0x0:
	case 0x1:
		OP  = MUL;
		if(op&1)
			OP2 = FLAG;
		else
			OP2 = NOP;
		break;
	case 0x2:
	case 0x3:
		OP  = MLA;
		if(op&1)
			OP2 = FLAG;
		else
			OP2 = NOP;
		break;
	case 0x4:
		OP  = UMAAL;
		OP2 = NOP;
		break;
	case 0x6:
		OP  = MLS;
		OP2 = NOP;
		break;
	case 0x8:
	case 0x9:
		OP  = UMULL;
		if(op&1)
			OP2 = FLAG;
		else
			OP2 = NOP;
		break;
	case 0xa:
	case 0xb:
		OP  = UMLAL;
		if(op&1)
			OP2 = FLAG;
		else
			OP2 = NOP;
		break;
	case 0xc:
	case 0xd:
		OP  = SMULL;
		if(op&1)
			OP2 = FLAG;
		else
			OP2 = NOP;
		break;
	case 0xe:
	case 0xf:
		OP  = SMLAL;
		if(op&1)
			OP2 = FLAG;
		else
			OP2 = NOP;
		break;
	default:
		dec_undef();
		break;
	}
}

static void dec_sync(void)
{
	rs1 = (inst_reg>>16)&0xf;
	rs2 = (inst_reg>> 0)&0xf;
	rd  = (inst_reg>>12)&0xf;
	op  = (inst_reg>>20)&0xf;
	if(op<8){
		if(op==0){
			OP  = NOP;
			OP2 = SWP;
			msize = WORD;
		}
		else if (op==4){
			OP  = NOP;
			OP2 = SWP;
			msize = BYTE;
		}
		else{
			dec_undef();
		}
		return;
	}

	if(op&1){
		OP  = NOP;
		OP2 = LDREX;
	}
	else{
		OP  = NOP;
		OP2 = STREX;
	}
	switch((op>>1)&3){
	case 0x0:
		msize = WORD;
		break;
	case 0x1:
		msize = DWORD;
		break;
	case 0x2:
		msize = BYTE;
		break;
	default:
		msize = HWORD;
		break;
	}
}

static void dec_ext_ldst(void)
{
	op1 = (inst_reg>>20)&0x5;
	op2 = (inst_reg>> 5)&0x3;
	rs1 = (inst_reg>>16)&0xf;
	madd   = (inst_reg & (1<<23))? 1 : 0;
	mindex = (inst_reg & (1<<24))? 1 : 0;
	mwback = (inst_reg & (1<<21))? 1 : 0;
	mwback |= (mindex==0);
	immd = (((inst_reg>> 8)&0xf)<<4)|(inst_reg & 0xf);
	switch(op2){
	case 0x1:
		if(op1==0){
			rs2 = (inst_reg>> 0)&0xf;
			rs3 = (inst_reg>>12)&0xf;
			rd  = NONE;
			OP  = EADDR;
			OP2 = ST;
			msize = HWORD;
			msex  = 0;
		}
		else if(op1==1){
			rs2 = (inst_reg>> 0)&0xf;
			rd  = (inst_reg>>12)&0xf;
			rs3 = NONE;
			OP  = EADDR;
			OP2 = LD;
			msize = HWORD;
			msex  = 0;
		}
		else if(op1==4){
			rs2 = IMMD;
			rs3 = (inst_reg>>12)&0xf;
			rd  = NONE;
			OP  = EADDR;
			OP2 = ST;
			msize = HWORD;
			msex  = 0;
		}
		else{
			rs2 = IMMD;
			rd  = (inst_reg>>12)&0xf;
			rs3 = NONE;
			OP  = EADDR;
			OP2 = LD;
			msize = HWORD;
			msex  = 0;
		}	
		break;
	case 0x2:
		if(op1==0){
			rs2 = (inst_reg>> 0)&0xf;
			rd  = (inst_reg>>12)&0xf;
			rs3 = NONE;
			OP  = EADDR;
			OP2 = LD;
			msize = DWORD;
			msex  = 0;
		}
		else if(op1==1){
			rs2 = (inst_reg>> 0)&0xf;
			rd  = (inst_reg>>12)&0xf;
			rs3 = NONE;
			OP  = EADDR;
			OP2 = LD;
			msize = BYTE;
			msex  = 1;
		}
		else if(op1==4){
			rs2 = IMMD;
			rd  = (inst_reg>>12)&0xf;
			rs3 = NONE;
			OP  = EADDR;
			OP2 = LD;
			msize = DWORD;
			msex  = 0;
		}
		else{
			rs2 = IMMD;
			rd  = (inst_reg>>12)&0xf;
			rs3 = NONE;
			OP  = EADDR;
			OP2 = LD;
			msize = BYTE;
			msex  = 1;
		}	
		break;
	case 0x3:
		if(op1==0){
			rs2 = (inst_reg>> 0)&0xf;
			rs3 = (inst_reg>>12)&0xf;
			rd  = NONE;
			OP  = EADDR;
			OP2 = ST;
			msize = DWORD;
			msex  = 0;
		}
		else if(op1==1){
			rs2 = (inst_reg>> 0)&0xf;
			rd  = (inst_reg>>12)&0xf;
			rs3 = NONE;
			OP  = EADDR;
			OP2 = LD;
			msize = HWORD;
			msex  = 1;
		}
		else if(op1==4){
			rs2 = IMMD;
			rs3 = (inst_reg>>12)&0xf;
			rd  = NONE;
			OP  = EADDR;
			OP2 = ST;
			msize = DWORD;
			msex  = 0;
		}
		else{
			rs2 = IMMD;
			rd  = (inst_reg>>12)&0xf;
			rs3 = NONE;
			OP  = EADDR;
			OP2 = LD;
			msize = HWORD;
			msex  = 1;
		}	
		break;
	default:
		dec_undef();
		break;
	}
}

static void dec_ext_ldst_unpriv(void)
{
	op  = (inst_reg>>20)&1;
	op2 = (inst_reg>> 5)&3;
	rs1 = (inst_reg>>20)&0xf;
	madd   = (inst_reg & (1<<23))? 1 : 0;
	mindex = 1;
	mwback = 1;
	immd = (((inst_reg>> 8)&0xf)<<4)|(inst_reg & 0xf);
	switch(op2){
	case 0x1:
		if(op==0){
			if(inst_reg & (1<<22))
				rs2 = IMMD;
			else
				rs2 = (inst_reg & 0xf);
			rs3 = (inst_reg>>12)&0xf;
			rd  = NONE;
			OP  = EADDR;
			OP2 = ST;
			msize = HWORD;
			msex  = 0;
		}
		else{
			if(inst_reg & (1<<22))
				rs2 = IMMD;
			else
				rs2 = (inst_reg & 0xf);
			rd  = (inst_reg>>12)&0xf;
			rs3 = NONE;
			OP  = EADDR;
			OP2 = LD;
			msize = HWORD;
			msex  = 0;
		}
		break;
	case 0x2:
		if(op==1){
			if(inst_reg & (1<<22))
				rs2 = IMMD;
			else
				rs2 = (inst_reg & 0xf);
			rd  = (inst_reg>>12)&0xf;
			rs3 = NONE;
			OP  = EADDR;
			OP2 = LD;
			msize = BYTE;
			msex  = 1;
		}
		else{
			dec_undef();
		}
		break;
	case 0x3:
		if(op==1){
			if(inst_reg & (1<<22))
				rs2 = IMMD;
			else
				rs2 = (inst_reg & 0xf);
			rd  = (inst_reg>>12)&0xf;
			rs3 = NONE;
			OP  = EADDR;
			OP2 = LD;
			msize = HWORD;
			msex  = 1;
		}
		else{
			dec_undef();
		}
		break;
	default:
		dec_undef();
		break;
	}
}

static void dec_movw(void)
{
	int imm4, imm12;
	imm4  = (inst_reg>>16)&0xf;
	imm12 = (inst_reg>> 0)&0xfff;
	immd  = (imm4<<12)|imm12;
	rs1   = NONE;
	rs2   = IMMD;
	rd    = (inst_reg>>12)&0xf;
	OP    = MOV;
	OP2   = NOP;
}

static void dec_movt(void)
{
	int imm4, imm12;
	imm4  = (inst_reg>>16)&0xf;
	imm12 = (inst_reg>> 0)&0xfff;
	immd  = (imm4<<12)|imm12;
	immd  = immd<<16;
	rs1   = (inst_reg>>12)&0xf;
	rs2   = IMMD;
	rd    = (inst_reg>>12)&0xf;
	OP    = OR;
	OP2   = NOP;
}

static void dec_msr(void)
{
	rs1 = NONE;
	rs2 = NONE;
	rs3 = NONE;
	rd  = NONE;
	OP  = NOP; /* not supported */
	OP2 = NOP; /* not supported */
}

static void dec_data_misc(void)
{
	op  = (inst_reg>>25)&0x1;
	op1 = (inst_reg>>20)&0x1f;
	op2 = (inst_reg>>4) &0xf;
	
	if(op){
		switch(op1){
		case 0x10:
			dec_movw();
			break;
		case 0x14:
			dec_movt();
			break;
		case 0x12:
		case 0x16:
			dec_msr();
			break;
		default:
			dec_data_proc_i();
			break;
		}
	}
	else{
                if(((op1&0x19)!=0x10)&&((op2&1)==0)){
			dec_data_proc_r();
		}
		else if(((op1&0x19)!=0x10)&&((op2&9)==1)){
			dec_data_proc_sr();
		}
		else if(((op1&0x19)==0x10)&&((op2&8)==0)){
			dec_misc();
		}
		else if(((op1&0x19)==0x10)&&((op2&9)==8)){
			dec_mulh();
		}
		else if(((op1&0x10)==0x00)&&(op2==9)){
			dec_mul();
		}
		else if(((op1&0x10)==0x10)&&(op2==9)){
			dec_sync();
		}
		else if(((op1&0x12)!=0x02)&&(op2==11)){
			dec_ext_ldst();
		}
		else if(((op1&0x12)!=0x02)&&((op2&13)==13)){
			dec_ext_ldst();
		}
		else if(((op1&0x13)==0x02)&&((op2&13)==13)){
			dec_ext_ldst();
		}
		else if(((op1&0x12)==0x02)&&(op2==11)){
			dec_ext_ldst_unpriv();
		}
		else if(((op1&0x13)==0x03)&&((op2&13)==13)){
			dec_ext_ldst_unpriv();
		}
		else{
			dec_undef();
		}
	}
}

static void dec_ldst_ub(void)
{
	int op1_xx0x0;
	int op1_0x010;
	int op1_xx0x1;
	int op1_0x011;
	int op1_xx1x0;
	int op1_0x110;
	int op1_xx1x1;
	int op1_0x111;
	int A,B;
	int A_0;
	int A_1_B_0;
	op1 = (inst_reg>>20)&0x1f;
	A   = (inst_reg>>25)&0x1;
	B   = (inst_reg>> 4)&0x1;
	rs1 = (inst_reg>>16)&0xf;
	madd   = (inst_reg & (1<<23))? 1 : 0;
	mindex = (inst_reg & (1<<24))? 1 : 0;
	mwback = (inst_reg & (1<<21))? 1 : 0;
	mwback |= (mindex==0);
	immd = (inst_reg & 0xfff);
	op1_xx0x0 = (op1&0x05)==0x00;
	op1_0x010 = (op1&0x17)==0x02;
	op1_xx0x1 = (op1&0x05)==0x01;
	op1_0x011 = (op1&0x17)==0x03;
	op1_xx1x0 = (op1&0x05)==0x04;
	op1_0x110 = (op1&0x17)==0x06;
	op1_xx1x1 = (op1&0x05)==0x05;
	op1_0x111 = (op1&0x17)==0x07;
	A_0 = A==0;
	A_1_B_0 = (A==1)&&(B==0);
	if(A_0 && op1_xx0x0 && !op1_0x010){ /* STR imm */
		rs2 = IMMD;
		rs3 = (inst_reg>>12)&0xf;
		rd  = NONE;
		OP  = EADDR;
		OP2 = ST;
		msize = WORD;
		msex = 0;
	}
	else if(A_1_B_0 && op1_xx0x0 && !op1_0x010){/* STR reg */
		immd = make_ishift(inst_reg&0xfff);
		rs2 = IMMD;
		rs3 = (inst_reg>>12)&0xf;
		rd  = NONE;
		OP  = EADDR;
		OP2 = ST;
		msize = WORD;
		msex = 0;
	}
	else if((A_0 && op1_0x010)||(A_1_B_0 && op1_0x010)){/* STRT */
		mindex = 0;
		mwback = 1;
		if(inst_reg & (1>>25)){/* reg */
			immd = make_ishift(inst_reg&0xfff);
		}
		else{/* imm */
		}
		rs2 = IMMD;
		rs3 = (inst_reg>>12)&0xf;
		rd  = NONE;
		OP  = EADDR;
		OP2 = ST;
		msize = WORD;
		msex = 0;
	}
	else if(A_0 && op1_xx0x1 && !op1_0x011){ /* LDR imm */
		rs2 = IMMD;
		rd  = (inst_reg>>12)&0xf;
		rs3 = NONE;
		OP  = EADDR;
		OP2 = LD;
		msize = WORD;
		msex = 0;
	}
	else if(A_1_B_0 && op1_xx0x1 && !op1_0x011){ /* LDR reg */
		immd = make_ishift(inst_reg&0xfff);
		rs2 = IMMD;
		rd  = (inst_reg>>12)&0xf;
		rs3 = NONE;
		OP  = EADDR;
		OP2 = LD;
		msize = WORD;
		msex = 0;
	}
	else if((A_0 && op1_0x011)||(A_1_B_0 && op1_0x011)){ /* LDRT */
		mindex = 0;
		mwback = 1;
		if(inst_reg & (1>>25)){/* reg */
			immd = make_ishift(inst_reg&0xfff);
		}
		else{/* imm */
		}
		rs2 = IMMD;
		rs3 = (inst_reg>>12)&0xf;
		rd  = NONE;
		OP  = EADDR;
		OP2 = LD;
		msize = WORD;
		msex = 0;
	}
	else if(A_0 && op1_xx1x0 && !op1_0x110){/* STRB imm */
		rs2 = IMMD;
		rs3 = (inst_reg>>12)&0xf;
		rd  = NONE;
		OP  = EADDR;
		OP2 = ST;
		msize = BYTE;
		msex = 0;
	}
	else if(A_1_B_0 && op1_xx1x0 && !op1_0x110){/* STRB reg */
		immd = make_ishift(inst_reg&0xfff);
		rs2 = IMMD;
		rs3 = (inst_reg>>12)&0xf;
		rd  = NONE;
		OP  = EADDR;
		OP2 = ST;
		msize = BYTE;
		msex = 0;
	}
	else if((A_0 && op1_0x110)||(A_1_B_0 && op1_0x110)){/* STRBT */
		mindex = 0;
		mwback = 1;
		if(inst_reg & (1>>25)){/* reg */
			immd = make_ishift(inst_reg&0xfff);
		}
		else{/* imm */
		}
		rs2 = IMMD;
		rs3 = (inst_reg>>12)&0xf;
		rd  = NONE;
		OP  = EADDR;
		OP2 = ST;
		msize = BYTE;
		msex = 0;
	}
	else if(A_0 && op1_xx1x1 && !op1_0x111){/* LDRB imm */
		rs2 = IMMD;
		rd  = (inst_reg>>12)&0xf;
		rs3 = NONE;
		OP  = EADDR;
		OP2 = LD;
		msize = BYTE;
		msex = 0;
	}
	else if(A_1_B_0 && op1_xx1x1 && !op1_0x111){/* LDRB reg */
		immd = make_ishift(inst_reg&0xfff);
		rs2 = IMMD;
		rd  = (inst_reg>>12)&0xf;
		rs3 = NONE;
		OP  = EADDR;
		OP2 = LD;
		msize = BYTE;
		msex = 0;
	}
	else if((A_0 && op1_0x111)||(A_1_B_0 && op1_0x111)){/* LDRBT */
		mindex = 0;
		mwback = 1;
		if(inst_reg & (1>>25)){/* reg */
			immd = make_ishift(inst_reg&0xfff);
		}
		else{/* imm */
		}
		rs2 = IMMD;
		rs3 = (inst_reg>>12)&0xf;
		rd  = NONE;
		OP  = EADDR;
		OP2 = LD;
		msize = BYTE;
		msex = 0;
	}
	else{
		dec_undef();
	}
}

static void dec_pack(void)
{
  int A;
  op1 = (inst_reg>>20)&7;
  op2 = (inst_reg>> 5)&7;
  rd  = (inst_reg>>12)&0xf;
  A   = (inst_reg>>16)&0xf;
  rs1 = NONE;
  rs2 = (inst_reg)&0xf;
  mrot = (inst_reg>>10)&0x3;
  if((op1==2)&&(op2==3)&&(A==0xf)){/*SXTB */
    OP  = SEX;
    OP2 = NOP;
    msize = BYTE;
    return;
  }
  if((op1==3)&&(op2==3)&&(A==0xf)){/*SXTH */
    OP  = SEX;
    OP2 = NOP;
    msize = HWORD;
    return;
  }
  if((op1==6)&&(op2==3)&&(A==0xf)){/*UXTB */
    OP  = ZEX;
    OP2 = NOP;
    msize = BYTE;
    return;
  }
  if((op1==7)&&(op2==3)&&(A==0xf)){/*UXTH */
    OP  = ZEX;
    OP2 = NOP;
    msize = HWORD;
    return;
  }
  dec_undef();
}


static void dec_media(void)
{
	op  = (inst_reg>>20)&0x1f;
	op2 = (inst_reg>> 5)&0x7;
	rd  = (inst_reg>>16)&0xf;
	rs1 = (inst_reg>> 0)&0xf;
	rs2 = (inst_reg>> 8)&0xf;
	if(op==0x11){
		OP  = SDIV;
		OP2 = NONE;
		return;
	}
	if(op==0x13){
		OP  = UDIV;
		OP2 = NONE;
		return;
	}
	if(((op>>3)&3)==1){
	  dec_pack();
	  return;
	}
	if(((op&0x1e)==0x1a)&&((op2&3)==2)){/* SBFX */
	  mlsb = (inst_reg>>7)&0x1f;
	  mwidth=((inst_reg>>16)&0x1f)+1;
	  rd = (inst_reg>>12)&0xf;
	  OP = SBFX;
	  OP2 = NOP;
	  return;
	}
	if(((op&0x1e)==0x1c)&&((op2&3)==0)){/* BFC BFI */
	  if(rs1==15){/* BFC */
	    mlsb = (inst_reg>>7)&0x1f;
	    mwidth=((inst_reg>>16)&0x1f)-mlsb+1;
	    rd = (inst_reg>>12)&0xf;
	    OP = BFC;
	    OP2 = NOP;
	  }
	  else{/* BFI */
	    mlsb = (inst_reg>>7)&0x1f;
	    mwidth=((inst_reg>>16)&0x1f)-mlsb+1;
	    rd = (inst_reg>>12)&0xf;
	    OP = BFI;
	    OP2 = NOP;
	  }
	  return;
	}
	if(((op&0x1e)==0x1e)&&((op2&3)==2)){/* UBFX */
	  mlsb = (inst_reg>>7)&0x1f;
	  mwidth=((inst_reg>>16)&0x1f)+1;
	  rd = (inst_reg>>12)&0xf;
	  OP = UBFX;
	  OP2 = NOP;
	  return;
	}


	dec_undef();
}

static void dec_branch_block()
{
	int R;
	op  = (inst_reg>>20)&0x3f;
	rs1 = (inst_reg>>16)&0xf;
	R   = (inst_reg>>15)&0x1;
	if((op&0x30)==0x20){/* Branch */
		immd = (inst_reg &0xffffff);
		if(immd & (1<<23))
			immd = immd | 0xff000000;
		immd = immd<<2;
		rs1 = NONE;
		rs2 = NONE;
		rs3 = NONE;
		rd  = NONE;
		OP  = NOP;
		OP2 = JMP;
		return;
	}
	if((op&0x30)==0x30){/* Branch with Link */
		immd = (inst_reg &0xffffff);
		if(immd & (1<<23))
			immd = immd | 0xff000000;
		immd = immd<<2;
		rs1 = NONE;
		rs2 = NEXT_PC;
		rs3 = NONE;
		rd  = 14;
		OP  = MOV;
		OP2 = JMP;
		return;
	}
	rs1 = (inst_reg>>16)&0xf;
	mlist = inst_reg&0xffff;
	
	switch((inst_reg>>23)&0x3){
	case 0:
		amode = DA;
		word_hi = 1;
		mincr = 0;
		break;
	case 1:
		amode = IA;
		word_hi = 0;
		mincr = 1;
		break;
	case 2:
		amode = DB;
		word_hi = 0;
		mincr = 0;
		break;
	case 3:
		amode = IB;
		word_hi = 1;
		mincr = 1;
		break;
	}
	mwback = (inst_reg>>21)&1;
	muser  = (inst_reg>>22)&1;
	mexret = (inst_reg>>15)&1;
	mexret = mexret & muser;
	if(op&1)
		OP = LDM;
	else
		OP = STM;
	OP2 = NOP;	
}

static void dec_copro()
{
	dec_undef();
}

void decode(void)
{
  cond    = (inst_reg>>28)&0xf;
  opcode  = (inst_reg>>25)&0x7;
  opcode2 = (inst_reg>>4)&0x1;

  if(cond==0xf){
    dec_uncond();
    return;
  }

  switch(opcode){
  case 0x0:
  case 0x1:
    dec_data_misc();
    break;
  case 0x2:
    dec_ldst_ub();
    break;
  case 0x3:
    if(opcode2)
      dec_media();
    else
      dec_ldst_ub();
    break;
  case 0x4:
  case 0x5:
    dec_branch_block();
    break;
  default:
    dec_copro();
    break;
  }
}
