#include <stdio.h>
#include <string.h>

#include "define.h"

static int cond;
static int rs1;
static int rs2;
static int rs3;
static int rd;
static int OP;
static int OP2;
static int msize;
static int msex;
static int n_high;
static int m_high;
static int madd;
static int mindex;
static int mwback;
static char* amode;
static int muser;
static int mexret;
static int immd;
static int mlist;
static int inst_reg;
static int PC;
static int opcode;
static int opcode2;
static int op;
static int op1;
static int op2;
static int mrot;
static int mlsb;
static int mwidth;

int ISFT32(int src, int sa, int* n, int* z, int* c, int* v);

static char* cond_name[]={
	"eq",
	"ne",
	"cs",
	"cc",
	"mi",
	"pl",
	"vs",
	"vc",
	"hi",
	"ls",
	"ge",
	"lt",
	"gt",
	"le",
	"",
	"",
};

static void dis_undef(void)
{
	printf(".word\t0x%08x\n",inst_reg);
}

static void dis_uncond(void)
{
	op1 = (inst_reg>>20)&0xff;
	rs1 = (inst_reg>>16)&0xf;
	dis_undef(); /* not supported */
}

static char* make_op(op)
{
	switch(op){
	case 0x0:
		return "and";
	case 0x1:
		return "xor";
	case 0x2:
		return "sub";
	case 0x3:
		return "rsb";
	case 0x4:
		return "add";
	case 0x5:
		return "adc";
	case 0x6:
		return "sbc";
	case 0x7:
		return "rsc";
	case 0x8:
		return "tst";
	case 0x9:
		return "teq";
	case 0xa:
		return "cmp";
	case 0xb:
	  return "cmn";
	case 0xc:
		return "or";
	case 0xd:
		return "mov";
	case 0xe:
		return "bic";
	case 0xf:
		return "mvn";
	}
}

static char rlist[2+16*3*2];

static char* make_rlist(int mlist)
{
	char rnam[8];
	int i;
	int rex=0;
	rlist[0]=0;
	strcpy(rlist,"{");
	for(i=0;i<16;i++){
		if(mlist&(1<<i)){
			if(i==0)
				sprintf(rnam,"r%d\0",i);
			else if(rex==0)
				sprintf(rnam,"r%d\0",i);
			else
			  sprintf(rnam,",r%d\0",i);
			strcat(rlist,rnam);
			rex |= 1;
		}
	}
	strcat(rlist,"}");
	return rlist;
}

static int make_immd12(int x)
{
	int dat,sa;
	dat = x & 0xff;
	sa  = (x>>8) & 0xf;
	sa  = sa<<1;
	return (dat>>sa)|(dat<<(32-sa));
}

static char is_reg[100];

static char* make_ishift(int x)
{
	int rm, st, sa;
	rm = x & 0xf;
	sa = (x>>7)&0x1f;
	st = (x>>5)&0x3;
	if(st==0){
	  if(sa==0){
	    sprintf(is_reg,"r%d\0",rm);
	  }
	  else{
	    sprintf(is_reg,"r%d, lsl #%d\0",rm, sa);
	  }
	}
	else if(st==1){
	  if(sa==0){
	    sprintf(is_reg,"r%d\0",rm);
	  }
	  else{
	    sprintf(is_reg,"r%d, lsr #%d\0",rm, sa);
	  }
	}
	else if(st==2){
	  if(sa==0){
	    sprintf(is_reg,"r%d\0",rm);
	  }
	  else{
	    sprintf(is_reg,"r%d, asr #%d\0",rm, sa);
	  }
	}
	else{
	  if(sa==0){
	    sprintf(is_reg,"r%d rrx\0",rm);
	  }
	  else{
	    sprintf(is_reg,"r%d, ror #%d\0",rm, sa);
	  }
	}
	return is_reg;
}

static char* make_rshift(int x)
{
	int rm, rs, st, sa;
	rm = x & 0xf;
	rs = (x>>8)&0xf;
	st = (x>>5)&0x3;
	if(st==0){
		sprintf(is_reg,"r%d, lsl r%d\0",rm, rs);
	}
	else if(st==1){
		sprintf(is_reg,"r%d, lsr r%d\0",rm, rs);
	}
	else if(st==2){
		sprintf(is_reg,"r%d, asr r%d\0",rm, rs);
	}
	else{
		sprintf(is_reg,"r%d, ror r%d\0",rm, rs);
	}
	return is_reg;
}

static void dis_data_proc_i(void)
{
	int op_tmp;
	int no_s;
	int is_mov;
	immd = make_immd12(inst_reg&0xfff);
	rs1  = (inst_reg>>16)&0xf;
	rs2  = IMMD;
	rd   = (inst_reg>>12)&0xf;
	op_tmp = (inst_reg>>21)&0xf;
	no_s = ((op_tmp>=8)&&(op_tmp<=11));
	is_mov = ((op_tmp==0xd)||(op_tmp==0xf));
	if((inst_reg>>20)&0x1){
	  if(no_s)
	    printf("%s%s\tr%d,#0x%08x\n",make_op(op_tmp),cond_name[cond],rs1,immd);
	  else if(is_mov)
	    printf("%s%s\tr%d,#0x%08x\n",make_op(op_tmp),cond_name[cond],rd,immd);
	  else
	    printf("%ss%s\tr%d,r%d,#0x%08x\n",make_op(op_tmp),cond_name[cond],rd,rs1,immd);
	}
	else{
	  if(is_mov)
	    printf("%s%s\tr%d,#0x%08x\n",make_op(op_tmp),cond_name[cond],rd,immd);
	  else
	    printf("%s%s\tr%d,r%d,#0x%08x\n",make_op(op_tmp),cond_name[cond],rd,rs1,immd);
	}
}

static void dis_data_proc_r(void)
{
	int op_tmp;
	int no_s;
	int is_mov;
	rs1  = (inst_reg>>16)&0xf;
	rs2  = IMMD;
	rd   = (inst_reg>>12)&0xf;
	op_tmp = (inst_reg>>21)&0xf;
	no_s = ((op_tmp>=8)&&(op_tmp<=11));
	is_mov = ((op_tmp==0xd)||(op_tmp==0xf));
	if((inst_reg>>20)&0x1){
	  if(no_s)
	    printf("%s%s\tr%d,%s\n",make_op(op_tmp),cond_name[cond],rs1,make_ishift(inst_reg&0xfff));
	  else if(is_mov)
	    printf("%s%s\tr%d,%s\n",make_op(op_tmp),cond_name[cond],rd,make_ishift(inst_reg&0xfff));
	  else
	    printf("%ss%s\tr%d,r%d,%s\n",make_op(op_tmp),cond_name[cond],rd,rs1,make_ishift(inst_reg&0xfff));
	}
	else{
	  if(is_mov)
	    printf("%s%s\tr%d,%s\n",make_op(op_tmp),cond_name[cond],rd,make_ishift(inst_reg&0xfff));
	  else
	    printf("%s%s\tr%d,r%d,%s\n",make_op(op_tmp),cond_name[cond],rd,rs1,make_ishift(inst_reg&0xfff));
	}
}

static void dis_data_proc_sr(void)
{
	int op_tmp;
	int no_s;
	rs1  = (inst_reg>>16)&0xf;
	rs2  = IMMD;
	rd   = (inst_reg>>12)&0xf;
	op_tmp = (inst_reg>>21)&0xf;
	no_s = ((op_tmp>=8)&&(op_tmp<=11));
	if((inst_reg>>20)&0x1){
		if(no_s)
			printf("%s%s\tr%d,r%d,%s\n",make_op(op_tmp),cond_name[cond],rd,rs1,make_rshift(inst_reg&0xfff));
		else
			printf("%ss%s\tr%d,r%d,%s\n",make_op(op_tmp),cond_name[cond],rd,rs1,make_rshift(inst_reg&0xfff));
	}
	else{
		printf("%s%s\tr%d,r%d,%s\n",make_op(op_tmp),cond_name[cond],rd,rs1,make_rshift(inst_reg&0xfff));
	}
}

static void dis_misc(void)
{
	op  = (inst_reg>>21)&0x3;
	op1 = (inst_reg>>16)&0xf;
	op2 = (inst_reg>> 4)&0x7;
	switch(op2){
	case 0x0:/* notsupported MRS/MSR */
		dis_undef();
		break;
	case 0x1:
		if(op==1){/* BX */
			rs1 = inst_reg&0xf;
			rs2 = NONE;
			rd  = NONE;
			printf("bx%s\tr%d\n",cond_name[cond],rs1);
		}
		else if(op==3){/* CLZ */
			rs1 = inst_reg&0xf;
			rs2 = NONE;
			rd  = (inst_reg>>12)&0xf;
			printf("clz%s\tr%d,r%d\n",cond_name[cond],rd,rs1);
		}
		else{
			dis_undef();
		}
		break;
	case 0x2:/* not supported BXJ */
		dis_undef();
		break;
	case 0x3:/* BLX */
		rs1 = inst_reg&0xf;
		printf("blx%s\tr%d\n",cond_name[cond],rs1);
		break;
	case 0x5:/* SAT ADD/SUB */
		dis_undef();
		break;
	case 0x6:/* not supported ERET */
		dis_undef();
		break;
	case 0x7:/* BKPT/HVC/SMC */
		dis_undef();
		break;
	default:
		dis_undef();
		break;
	}
}

static void dis_mulh(void)
{
	op  = (inst_reg>>5)&0x1;
	op1 = (inst_reg>>21)&0x3;
	rd  = (inst_reg>>16)&0xf;
	rs1 = (inst_reg>> 0)&0xf;
	rs2 = (inst_reg>> 8)&0xf;
	rs3 = (inst_reg>>12)&0xf;
	n_high = (inst_reg>>5)&1;
	m_high = (inst_reg>>6)&0x1;
	switch(op1){
	case 0x0:
		printf("smla%c%c%s\tr%d,r%d,r%d,r%d\n",(n_high)? 't' : 'b', (m_high)? 't' : 'b',cond_name[cond],rd,rs1,rs2,rs3);
		break;	
	case 0x1:
		if(op==0)
			printf("smlaw%c%s\tr%d,r%d,r%d,r%d\n", (m_high)? 't' : 'b',cond_name[cond],rd,rs1,rs2,rs3);
		else
			printf("smulw%c%s\tr%d,r%d,r%d\n", (m_high)? 't' : 'b',cond_name[cond],rd,rs1,rs2);
		break;	
	case 0x2:
		printf("smlal%c%c%s\tr%d,r%d,r%d,r%d\n",(n_high)? 't' : 'b', (m_high)? 't' : 'b',cond_name[cond],rs3,rd,rs1,rs2);
		break;	
	case 0x3:
		printf("smul%c%c%s\tr%d,r%d,r%d\n", (n_high)? 't' : 'b',(m_high)? 't' : 'b',cond_name[cond],rd,rs1,rs2);
		break;	
	}
}

static void dis_mul(void)
{
	op  = (inst_reg>>20)&0xf;
	rd  = (inst_reg>>16)&0xf;
	rs1 = (inst_reg>> 0)&0xf;
	rs2 = (inst_reg>> 8)&0xf;
	rs3 = (inst_reg>>12)&0xf;
	switch(op){
	case 0x0:
	case 0x1:
		if(op&1)
			printf("muls%s\tr%d,r%d,r%d\n",cond_name[cond],rd,rs1,rs2);
		else
			printf("mul%s\tr%d,r%d,r%d\n",cond_name[cond],rd,rs1,rs2);
		break;
	case 0x2:
	case 0x3:
		if(op&1)
			printf("mlas%s\tr%d,r%d,r%d,r%d\n",cond_name[cond],rd,rs1,rs2,rs3);
		else
			printf("mla%s\tr%d,r%d,r%d,r%d\n",cond_name[cond],rd,rs1,rs2,rs3);
		break;
	case 0x4:
		printf("umaal%s\tr%d,r%d,r%d,r%d\n",cond_name[cond],rs3,rd,rs1,rs2);
		break;
	case 0x6:
		printf("mls%s\tr%d,r%d,r%d,r%d\n",cond_name[cond],rd,rs1,rs2,rs3);
		OP2 = NOP;
		break;
	case 0x8:
	case 0x9:
		if(op&1)
			printf("umulls%s\tr%d,r%d,r%d,r%d\n",cond_name[cond],rs3,rd,rs1,rs2);
		else
			printf("umull%s\tr%d,r%d,r%d,r%d\n",cond_name[cond],rs3,rd,rs1,rs2);
		break;
	case 0xa:
	case 0xb:
		if(op&1)
			printf("umlals%s\tr%d,r%d,r%d,r%d\n",cond_name[cond],rs3,rd,rs1,rs2);
		else
			printf("umlal%s\tr%d,r%d,r%d,r%d\n",cond_name[cond],rs3,rd,rs1,rs2);
		break;
	case 0xc:
	case 0xd:
		if(op&1)
			printf("smulls%s\tr%d,r%d,r%d,r%d\n",cond_name[cond],rs3,rd,rs1,rs2);
		else
			printf("smull%s\tr%d,r%d,r%d,r%d\n",cond_name[cond],rs3,rd,rs1,rs2);
		break;
	case 0xe:
	case 0xf:
		if(op&1)
			printf("smlals%s\tr%d,r%d,r%d,r%d\n",cond_name[cond],rs3,rd,rs1,rs2);
		else
			printf("smlal%s\tr%d,r%d,r%d,r%d\n",cond_name[cond],rs3,rd,rs1,rs2);
		break;
	default:
		dis_undef();
		break;
	}
}

static void dis_sync(void)
{
	rs1 = (inst_reg>>16)&0xf;
	rs2 = (inst_reg>> 0)&0xf;
	rd  = (inst_reg>>12)&0xf;
	op  = (inst_reg>>20)&0xf;
	if(op<8){
		if(op==0){
			printf("swp%s\tr%d,r%d,[r%d]\n",cond_name[cond],rd,rs2,rs1);
		}
		else if (op==4){
			printf("swpb%s\tr%d,r%d,[r%d]\n",cond_name[cond],rd,rs2,rs1);
		}
		else{
			dis_undef();
		}
		return;
	}

	if(op&1){
		switch((op>>1)&3){
		case 0x0:
			printf("ldrex%s\tr%d,[r%d]\n",cond_name[cond],rd,rs1);
			break;
		case 0x1:
			printf("ldrexd%s\tr%d,r%d,[r%d]\n",cond_name[cond],rd,(rd+1)&0xf,rs1);
			break;
		case 0x2:
			printf("ldrexb%s\tr%d,[r%d]\n",cond_name[cond],rd,rs1);
			break;
		default:
			printf("ldrexh%s\tr%d,[r%d]\n",cond_name[cond],rd,rs1);
			break;
		}
	}
	else{
		switch((op>>1)&3){
		case 0x0:
			printf("strex%s\tr%d,r%d,[r%d]\n",cond_name[cond],rd,rs2,rs1);
			break;
		case 0x1:
			printf("strexd%s\tr%d,r%d,r%d,[r%d]\n",cond_name[cond],rd,rs2,(rs2+1)&0xf,rs1);
			break;
		case 0x2:
			printf("strexb%s\tr%d,r%d,[r%d]\n",cond_name[cond],rd,rs2,rs1);
			break;
		default:
			printf("strexh%s\tr%d,r%d,[r%d]\n",cond_name[cond],rd,rs2,rs1);
			break;
		}
	}
}

static void dis_ext_ldst(void)
{
	op1 = (inst_reg>>20)&0x5;
	op2 = (inst_reg>> 5)&0x3;
	rs1 = (inst_reg>>16)&0xf;
	madd   = (inst_reg & (1<<23))? 1 : 0;
	mindex = (inst_reg & (1<<24))? 1 : 0;
	mwback = (inst_reg & (1<<21))? 1 : 0;
	mwback |= (mindex==0);
	immd = (((inst_reg>> 8)&0xf)<<4)|(inst_reg & 0xf);
	switch(op2){
	case 0x1:
		if(op1==0){
			rs2 = (inst_reg>> 0)&0xf;
			rs3 = (inst_reg>>12)&0xf;
			rd  = NONE;
			if(mindex && !mwback){
				printf("strh%s\tr%d,[r%d,%cr%d]\n",cond_name[cond],rs3,rs1,(madd)? '+' : '-',rs2);
			} else if(mindex && mwback){
				printf("strh%s\tr%d,[r%d,%cr%d]!\n",cond_name[cond],rs3,rs1,(madd)? '+' : '-',rs2);
			} else if(!mindex && mwback){
				printf("strh%s\tr%d,[r%d],%cr%d\n",cond_name[cond],rs3,rs1,(madd)? '+' : '-',rs2);
			}
			else{
				dis_undef();
			}
		}
		else if(op1==1){
			rs2 = (inst_reg>> 0)&0xf;
			rd  = (inst_reg>>12)&0xf;
			if(mindex && !mwback){
				printf("ldrh%s\tr%d,[r%d,%cr%d]\n",cond_name[cond],rd,rs1,(madd)? '+' : '-',rs2);
			} else if(mindex && mwback){
				printf("ldrh%s\tr%d,[r%d,%cr%d]!\n",cond_name[cond],rd,rs1,(madd)? '+' : '-',rs2);
			} else if(!mindex && mwback){
				printf("ldrh%s\tr%d,[r%d],%cr%d\n",cond_name[cond],rd,rs1,(madd)? '+' : '-',rs2);
			}
			else{
				dis_undef();
			}
		}
		else if(op1==4){
			rs2 = IMMD;
			rs3 = (inst_reg>>12)&0xf;
			rd  = NONE;
			if(mindex && !mwback){
				printf("strh%s\tr%d,[r%d,#%c0x%02x]\n",cond_name[cond],rs3,rs1,(madd)? '+' : '-',immd);
			} else if(mindex && mwback){
				printf("strh%s\tr%d,[r%d,#%c0x%02x]!\n",cond_name[cond],rs3,rs1,(madd)? '+' : '-',immd);
			} else if(!mindex && mwback){
				printf("strh%s\tr%d,[r%d],#%c0x%02x\n",cond_name[cond],rs3,rs1,(madd)? '+' : '-',immd);
			}
			else{
				dis_undef();
			}
		}
		else{
			rs2 = IMMD;
			rd  = (inst_reg>>12)&0xf;
			rs3 = NONE;
			if(mindex && !mwback){
				printf("ldrh%s\tr%d,[r%d,#%c0x%02x]\n",cond_name[cond],rd,rs1,(madd)? '+' : '-',immd);
			} else if(mindex && mwback){
				printf("ldrh%s\tr%d,[r%d,#%c0x%02x]!\n",cond_name[cond],rd,rs1,(madd)? '+' : '-',immd);
			} else if(!mindex && mwback){
				printf("ldrh%s\tr%d,[r%d],#%c0x%02x\n",cond_name[cond],rd,rs1,(madd)? '+' : '-',immd);
			}
			else{
				dis_undef();
			}
		}	
		break;
	case 0x2:
		if(op1==0){
			rs2 = (inst_reg>> 0)&0xf;
			rd  = (inst_reg>>12)&0xf;
			rs3 = NONE;
			if(mindex && !mwback){
				printf("ldrd%s\tr%d,r%d,[r%d,%cr%d]\n",cond_name[cond],rd,(rd+1)&0xf,rs1,(madd)? '+' : '-',rs2);
			} else if(mindex && mwback){
				printf("ldrd%s\tr%d,r%d,[r%d,%cr%d]!\n",cond_name[cond],rd,(rd+1)&0xf,rs1,(madd)? '+' : '-',rs2);
			} else if(!mindex && mwback){
				printf("ldrd%s\tr%d,r%d,[r%d],%cr%d\n",cond_name[cond],rd,(rd*1)&0xf,rs1,(madd)? '+' : '-',rs2);
			}
			else{
				dis_undef();
			}
		}
		else if(op1==1){
			rs2 = (inst_reg>> 0)&0xf;
			rd  = (inst_reg>>12)&0xf;
			rs3 = NONE;
			if(mindex && !mwback){
				printf("ldrsb%s\tr%d,[r%d,%cr%d]\n",cond_name[cond],rd,rs1,(madd)? '+' : '-',rs2);
			} else if(mindex && mwback){
				printf("ldrsb%s\tr%d,[r%d,%cr%d]!\n",cond_name[cond],rd,rs1,(madd)? '+' : '-',rs2);
			} else if(!mindex && mwback){
				printf("ldrsb%s\tr%d,[r%d],%cr%d\n",cond_name[cond],rd,rs1,(madd)? '+' : '-',rs2);
			}
			else{
				dis_undef();
			}
		}
		else if(op1==4){
			rs2 = IMMD;
			rd  = (inst_reg>>12)&0xf;
			rs3 = NONE;
			if(mindex && !mwback){
				printf("ldrd%s\tr%d,r%d,[r%d,#%c0x%02x]\n",cond_name[cond],rd,(rd+1)&0xf,rs1,(madd)? '+' : '-',immd);
			} else if(mindex && mwback){
				printf("ldrd%s\tr%d,r%d,[r%d,#%c0x%02x]!\n",cond_name[cond],rd,(rd+1)&0xf,rs1,(madd)? '+' : '-',immd);
			} else if(!mindex && mwback){
				printf("ldrd%s\tr%d,r%d,[r%d],#%c0x%02x\n",cond_name[cond],rd,(rd+1)&0xf,rs1,(madd)? '+' : '-',immd);
			}
			else{
				dis_undef();
			}
		}
		else{
			rs2 = IMMD;
			rd  = (inst_reg>>12)&0xf;
			rs3 = NONE;
			if(mindex && !mwback){
				printf("ldrsb%s\tr%d,[r%d,#%c0x%02x]\n",cond_name[cond],rd,rs1,(madd)? '+' : '-',immd);
			} else if(mindex && mwback){
				printf("ldrsb%s\tr%d,[r%d,#%c0x%02x]!\n",cond_name[cond],rd,rs1,(madd)? '+' : '-',immd);
			} else if(!mindex && mwback){
				printf("ldrsb%s\tr%d,[r%d],#%c0x%02x\n",cond_name[cond],rd,rs1,(madd)? '+' : '-',immd);
			}
			else{
				dis_undef();
			}
		}	
		break;
	case 0x3:
		if(op1==0){
			rs2 = (inst_reg>> 0)&0xf;
			rs3 = (inst_reg>>12)&0xf;
			rd  = NONE;
			if(mindex && !mwback){
				printf("strd%s\tr%d,r%d,[r%d,%cr%d]\n",cond_name[cond],rs3,(rs3+1)&0xf,rs1,(madd)? '+' : '-',rs2);
			} else if(mindex && mwback){
				printf("strd%s\tr%d,r%d,[r%d,%cr%d]!\n",cond_name[cond],rs3,(rs3+1)&0xf,rs1,(madd)? '+' : '-',rs2);
			} else if(!mindex && mwback){
				printf("strd%s\tr%d,r%d,[r%d],%cr%d\n",cond_name[cond],rs3,(rs3+1)&0xf,rs1,(madd)? '+' : '-',rs2);
			}
			else{
				dis_undef();
			}
		}
		else if(op1==1){
			rs2 = (inst_reg>> 0)&0xf;
			rd  = (inst_reg>>12)&0xf;
			rs3 = NONE;
			if(mindex && !mwback){
				printf("ldrsh%s\tr%d,[r%d,%cr%d]\n",cond_name[cond],rd,rs1,(madd)? '+' : '-',rs2);
			} else if(mindex && mwback){
				printf("ldrsh%s\tr%d,[r%d,%cr%d]!\n",cond_name[cond],rd,rs1,(madd)? '+' : '-',rs2);
			} else if(!mindex && mwback){
				printf("ldrsh%s\tr%d,[r%d],%cr%d\n",cond_name[cond],rd,rs1,(madd)? '+' : '-',rs2);
			}
			else{
				dis_undef();
			}
		}
		else if(op1==4){
			rs2 = IMMD;
			rs3 = (inst_reg>>12)&0xf;
			rd  = NONE;
			if(mindex && !mwback){
				printf("strd%s\tr%d,[r%d,#%c0x%02x]\n",cond_name[cond],rs3,(rs3+1)&0xf,rs1,(madd)? '+' : '-',immd);
			} else if(mindex && mwback){
				printf("strd%s\tr%d,[r%d,#%c0x%02x]!\n",cond_name[cond],rs3,(rs3+1)&0xf,rs1,(madd)? '+' : '-',immd);
			} else if(!mindex && mwback){
				printf("strd%s\tr%d,[r%d],#%c0x%02x\n",cond_name[cond],rs3,(rs3+1)&0xf,rs1,(madd)? '+' : '-',immd);
			}
			else{
				dis_undef();
			}
		}
		else{
			rs2 = IMMD;
			rd  = (inst_reg>>12)&0xf;
			rs3 = NONE;
			if(mindex && !mwback){
				printf("ldrsh%s\tr%d,[r%d,#%c0x%02x]\n",cond_name[cond],rd,rs1,(madd)? '+' : '-',immd);
			} else if(mindex && mwback){
				printf("ldrsh%s\tr%d,[r%d,#%c0x%02x]!\n",cond_name[cond],rd,rs1,(madd)? '+' : '-',immd);
			} else if(!mindex && mwback){
				printf("ldrsh%s\tr%d,[r%d],#%c0x%02x\n",cond_name[cond],rd,rs1,(madd)? '+' : '-',immd);
			}
			else{
				dis_undef();
			}
		}	
		break;
	default:
		dis_undef();
		break;
	}
}

static void dis_ext_ldst_unpriv(void)
{
	op  = (inst_reg>>20)&1;
	op2 = (inst_reg>> 5)&3;
	rs1 = (inst_reg>>20)&0xf;
	madd   = (inst_reg & (1<<23))? 1 : 0;
	mindex = 1;
	mwback = 1;
	immd = (((inst_reg>> 8)&0xf)<<4)|(inst_reg & 0xf);
	switch(op2){
	case 0x1:
		rs2 = (inst_reg & 0xf);
		rs3 = (inst_reg>>12)&0xf;
		if(op==0){
			if(inst_reg & (1<<22))
				printf("strht%s\tr%d,[r%d],#%c0x02x\n",cond_name[cond],rs3,rs1,(madd)? '+' : '-',immd);
			else
				printf("strht%s\tr%d,[r%d],%cr%d\n",cond_name[cond],rs3,rs1,(madd)? '+' : '-',rs2);
		}
		else{
			rs2 = (inst_reg & 0xf);
			rd  = (inst_reg>>12)&0xf;
			if(inst_reg & (1<<22))
				printf("ldrht%s\tr%d,[r%d],#%c0x02x\n",cond_name[cond],rd,rs1,(madd)? '+' : '-',immd);
			else
				printf("ldrht%s\tr%d,[r%d],%cr%d\n",cond_name[cond],rd,rs1,(madd)? '+' : '-',rs2);
		}
		break;
	case 0x2:
		rs2 = (inst_reg & 0xf);
		rd  = (inst_reg>>12)&0xf;
		if(op==1){
			if(inst_reg & (1<<22))
				printf("ldrsbt%s\tr%d,[r%d],#%c0x02x\n",cond_name[cond],rd,rs1,(madd)? '+' : '-',immd);
			else
				printf("ldrsbt%s\tr%d,[r%d],%cr%d\n",cond_name[cond],rd,rs1,(madd)? '+' : '-',rs2);
		}
		else{
			dis_undef();
		}
		break;
	case 0x3:
		rs2 = (inst_reg & 0xf);
		rd  = (inst_reg>>12)&0xf;
		if(op==1){
			if(inst_reg & (1<<22))
				printf("ldrsht%s\tr%d,[r%d],#%c0x02x\n",cond_name[cond],rd,rs1,(madd)? '+' : '-',immd);
			else
				printf("ldrsht%s\tr%d,[r%d],%cr%d\n",cond_name[cond],rd,rs1,(madd)? '+' : '-',rs2);
		}
		else{
			dis_undef();
		}
		break;
	default:
		dis_undef();
		break;
	}
}

static void dis_movw(void)
{
	int imm4, imm12;
	imm4  = (inst_reg>>16)&0xf;
	imm12 = (inst_reg>> 0)&0xfff;
	immd  = (imm4<<12)|imm12;
	rs1   = NONE;
	rs2   = IMMD;
	rd    = (inst_reg>>12)&0xf;
	printf("movw%s\tr%d,#0x%04x\n",cond_name[cond],rd,immd);
}

static void dis_movt(void)
{
	int imm4, imm12;
	imm4  = (inst_reg>>16)&0xf;
	imm12 = (inst_reg>> 0)&0xfff;
	immd  = (imm4<<12)|imm12;
	rs1   = (inst_reg>>12)&0xf;
	rs2   = IMMD;
	rd    = (inst_reg>>12)&0xf;
	printf("movt%s\tr%d,#0x%04x\n",cond_name[cond],rd,immd);
}

static void dis_msr(void)
{
	rs1 = NONE;
	rs2 = NONE;
	rs3 = NONE;
	rd  = NONE;
	printf(".word\t0x%08x\n",inst_reg);
}

static void dis_data_misc(void)
{
	op  = (inst_reg>>25)&0x1;
	op1 = (inst_reg>>20)&0x1f;
	op2 = (inst_reg>>4) &0xf;
	
	if(op){
		switch(op1){
		case 0x10:
			dis_movw();
			break;
		case 0x14:
			dis_movt();
			break;
		case 0x12:
		case 0x16:
			dis_msr();
			break;
		default:
			dis_data_proc_i();
			break;
		}
	}
	else{
		if(((op1&0x19)!=0x10)&&((op2&1)==0)){
			dis_data_proc_r();
		}
		else if(((op1&0x19)!=0x10)&&((op2&9)==1)){
			dis_data_proc_sr();
		}
		else if(((op1&0x19)==0x10)&&((op2&8)==0)){
			dis_misc();
		}
		else if(((op1&0x19)==0x10)&&((op2&9)==8)){
			dis_mulh();
		}
		else if(((op1&0x10)==0x00)&&(op2==9)){
			dis_mul();
		}
		else if(((op1&0x10)==0x10)&&(op2==9)){
			dis_sync();
		}
		else if(((op1&0x12)!=0x02)&&(op2==11)){
			dis_ext_ldst();
		}
		else if(((op1&0x12)!=0x02)&&((op2&13)==13)){
			dis_ext_ldst();
		}
		else if(((op1&0x13)==0x02)&&((op2&13)==13)){
			dis_ext_ldst();
		}
		else if(((op1&0x12)==0x02)&&(op2==11)){
			dis_ext_ldst_unpriv();
		}
		else if(((op1&0x13)==0x03)&&((op2&13)==13)){
			dis_ext_ldst_unpriv();
		}
		else{
			dis_undef();
		}
	}
}

static void dis_ldst_ub(void)
{
	int op1_xx0x0;
	int op1_0x010;
	int op1_xx0x1;
	int op1_0x011;
	int op1_xx1x0;
	int op1_0x110;
	int op1_xx1x1;
	int op1_0x111;
	int A,B;
	int A_0;
	int A_1_B_0;
	op1 = (inst_reg>>20)&0x1f;
	A   = (inst_reg>>25)&0x1;
	B   = (inst_reg>> 4)&0x1;
	rs1 = (inst_reg>>16)&0xf;
	madd   = (inst_reg & (1<<23))? 1 : 0;
	mindex = (inst_reg & (1<<24))? 1 : 0;
	mwback = (inst_reg & (1<<21))? 1 : 0;
	mwback |= (mindex==0);
	immd = (inst_reg & 0xfff);
	op1_xx0x0 = (op1&0x05)==0x00;
	op1_0x010 = (op1&0x17)==0x02;
	op1_xx0x1 = (op1&0x05)==0x01;
	op1_0x011 = (op1&0x17)==0x03;
	op1_xx1x0 = (op1&0x05)==0x04;
	op1_0x110 = (op1&0x17)==0x06;
	op1_xx1x1 = (op1&0x05)==0x05;
	op1_0x111 = (op1&0x17)==0x07;
	A_0 = A==0;
	A_1_B_0 = (A==1)&&(B==0);
	if(A_0 && op1_xx0x0 && !op1_0x010){ /* STR imm */
		rs2 = IMMD;
		rs3 = (inst_reg>>12)&0xf;
		rd  = NONE;
		if(mindex && !mwback){
			printf("str%s\tr%d,[r%d,#%c0x%03x]\n",cond_name[cond],rs3,rs1,(madd)? '+' : '-',immd);
		} else if(mindex && mwback){
			printf("str%s\tr%d,[r%d,#%c0x%03x]!\n",cond_name[cond],rs3,rs1,(madd)? '+' : '-',immd);
		} else if(!mindex && mwback){
			printf("str%s\tr%d,[r%d],#%c0x%03x\n",cond_name[cond],rs3,rs1,(madd)? '+' : '-',immd);
		}
		else{
			dis_undef();
		}
	}
	else if(A_1_B_0 && op1_xx0x0 && !op1_0x010){/* STR reg */
		rs2 = IMMD;
		rs3 = (inst_reg>>12)&0xf;
		rd  = NONE;
		if(mindex && !mwback){
			printf("str%s\tr%d,[r%d,%c%s]\n",cond_name[cond],rs3,rs1,(madd)? '+' : '-',make_ishift(inst_reg&0xfff));
		} else if(mindex && mwback){
			printf("str%s\tr%d,[r%d,%c%s]!\n",cond_name[cond],rs3,rs1,(madd)? '+' : '-',make_ishift(inst_reg&0xfff));
		} else if(!mindex && mwback){
			printf("str%s\tr%d,[r%d],%c%s\n",cond_name[cond],rs3,rs1,(madd)? '+' : '-',make_ishift(inst_reg&0xfff));
		}
		else{
			dis_undef();
		}
	}
	else if((A_0 && op1_0x010)||(A_1_B_0 && op1_0x010)){/* STRT */
		mindex = 0;
		mwback = 1;
		rs2 = IMMD;
		rs3 = (inst_reg>>12)&0xf;
		rd  = NONE;
		if(inst_reg & (1>>25)){/* reg */
			printf("strt%s\tr%d,[r%d],#%c%s\n",cond_name[cond],rs3,rs1,(madd)? '+' : '-',make_ishift(inst_reg&0xfff));
		}
		else{/* imm */
			printf("strt%s\tr%d,[r%d],#%c0x%03x\n",cond_name[cond],rs3,rs1,(madd)? '+' : '-',immd);
		}
	}
	else if(A_0 && op1_xx0x1 && !op1_0x011){ /* LDR imm */
		rs2 = IMMD;
		rd  = (inst_reg>>12)&0xf;
		rs3 = NONE;
		if(mindex && !mwback){
			printf("ldr%s\tr%d,[r%d,#%c0x%03x]\n",cond_name[cond],rd,rs1,(madd)? '+' : '-',immd);
		} else if(mindex && mwback){
			printf("ldr%s\tr%d,[r%d,#%c0x%03x]!\n",cond_name[cond],rd,rs1,(madd)? '+' : '-',immd);
		} else if(!mindex && mwback){
			printf("ldr%s\tr%d,[r%d],#%c0x%03x\n",cond_name[cond],rd,rs1,(madd)? '+' : '-',immd);
		}
		else{
			dis_undef();
		}
	}
	else if(A_1_B_0 && op1_xx0x1 && !op1_0x011){ /* LDR reg */
		rs2 = IMMD;
		rd  = (inst_reg>>12)&0xf;
		rs3 = NONE;
		if(mindex && !mwback){
			printf("ldr%s\tr%d,[r%d,%c%s]\n",cond_name[cond],rd,rs1,(madd)? '+' : '-',make_ishift(inst_reg&0xfff));
		} else if(mindex && mwback){
			printf("ldr%s\tr%d,[r%d,%c%s]!\n",cond_name[cond],rd,rs1,(madd)? '+' : '-',make_ishift(inst_reg&0xfff));
		} else if(!mindex && mwback){
			printf("ldr%s\tr%d,[r%d],%c%s\n",cond_name[cond],rd,rs1,(madd)? '+' : '-',make_ishift(inst_reg&0xfff));
		}
		else{
			dis_undef();
		}
	}
	else if((A_0 && op1_0x011)||(A_1_B_0 && op1_0x011)){ /* LDRT */
		mindex = 0;
		mwback = 1;
		rs2 = IMMD;
		rs3 = (inst_reg>>12)&0xf;
		rd  = NONE;
		if(inst_reg & (1>>25)){/* reg */
			printf("ldrt%s\tr%d,[r%d],#%c%s\n",cond_name[cond],rd,rs1,(madd)? '+' : '-',make_ishift(inst_reg&0xfff));
		}
		else{/* imm */
			printf("ldr%s\tr%d,[r%d],#%c0x%03x\n",cond_name[cond],rd,rs1,(madd)? '+' : '-',immd);
		}
	}
	else if(A_0 && op1_xx1x0 && !op1_0x110){/* STRB imm */
		rs2 = IMMD;
		rs3 = (inst_reg>>12)&0xf;
		rd  = NONE;
		if(mindex && !mwback){
			printf("strb%s\tr%d,[r%d,#%c0x%03x]\n",cond_name[cond],rs3,rs1,(madd)? '+' : '-',immd);
		} else if(mindex && mwback){
			printf("strb%s\tr%d,[r%d,#%c0x%03x]!\n",cond_name[cond],rs3,rs1,(madd)? '+' : '-',immd);
		} else if(!mindex && mwback){
			printf("strb%s\tr%d,[r%d],#%c0x%03x\n",cond_name[cond],rs3,rs1,(madd)? '+' : '-',immd);
		}
		else{
			dis_undef();
		}
	}
	else if(A_1_B_0 && op1_xx1x0 && !op1_0x110){/* STRB reg */
		rs2 = IMMD;
		rs3 = (inst_reg>>12)&0xf;
		rd  = NONE;
		if(mindex && !mwback){
			printf("strb%s\tr%d,[r%d,%c%s]\n",cond_name[cond],rs3,rs1,(madd)? '+' : '-',make_ishift(inst_reg&0xfff));
		} else if(mindex && mwback){
			printf("strb%s\tr%d,[r%d,%c%s]!\n",cond_name[cond],rs3,rs1,(madd)? '+' : '-',make_ishift(inst_reg&0xfff));
		} else if(!mindex && mwback){
			printf("strb%s\tr%d,[r%d],%c%s\n",cond_name[cond],rs3,rs1,(madd)? '+' : '-',make_ishift(inst_reg&0xfff));
		}
		else{
			dis_undef();
		}
	}
	else if((A_0 && op1_0x110)||(A_1_B_0 && op1_0x110)){/* STRBT */
		mindex = 0;
		mwback = 1;
		rs2 = IMMD;
		rs3 = (inst_reg>>12)&0xf;
		rd  = NONE;
		if(inst_reg & (1>>25)){/* reg */
			printf("strbt%s\tr%d,[r%d],%c%s\n",cond_name[cond],rs3,rs1,(madd)? '+' : '-',make_ishift(inst_reg&0xfff));
		}
		else{/* imm */
			printf("strbt%s\tr%d,[r%d],#%c0x%03x\n",cond_name[cond],rs3,rs1,(madd)? '+' : '-',immd);
		}
	}
	else if(A_0 && op1_xx1x1 && !op1_0x111){/* LDRB imm */
		rs2 = IMMD;
		rd  = (inst_reg>>12)&0xf;
		rs3 = NONE;
		if(mindex && !mwback){
			printf("ldrb%s\tr%d,[r%d,#%c0x%03x]\n",cond_name[cond],rd,rs1,(madd)? '+' : '-',immd);
		} else if(mindex && mwback){
			printf("ldrb%s\tr%d,[r%d,#%c0x%03x]!\n",cond_name[cond],rd,rs1,(madd)? '+' : '-',immd);
		} else if(!mindex && mwback){
			printf("ldrb%s\tr%d,[r%d],#%c0x%03x\n",cond_name[cond],rd,rs1,(madd)? '+' : '-',immd);
		}
		else{
			dis_undef();
		}
	}
	else if(A_1_B_0 && op1_xx1x1 && !op1_0x111){/* LDRB reg */
		rs2 = IMMD;
		rd  = (inst_reg>>12)&0xf;
		rs3 = NONE;
		if(mindex && !mwback){
			printf("ldrb%s\tr%d,[r%d,%c%s]\n",cond_name[cond],rs3,rs1,(madd)? '+' : '-',make_ishift(inst_reg&0xfff));
		} else if(mindex && mwback){
			printf("ldrb%s\tr%d,[r%d,%c%s]!\n",cond_name[cond],rs3,rs1,(madd)? '+' : '-',make_ishift(inst_reg&0xfff));
		} else if(!mindex && mwback){
			printf("ldrb%s\tr%d,[r%d],%c%s\n",cond_name[cond],rs3,rs1,(madd)? '+' : '-',make_ishift(inst_reg&0xfff));
		}
		else{
			dis_undef();
		}
	}
	else if((A_0 && op1_0x111)||(A_1_B_0 && op1_0x111)){/* LDRBT */
		mindex = 0;
		mwback = 1;
		rs2 = IMMD;
		rd  = (inst_reg>>12)&0xf;
		rs3 = NONE;
		if(inst_reg & (1>>25)){/* reg */
			printf("ldrbt%s\tr%d,[r%d],%c%s\n",cond_name[cond],rd,rs1,(madd)? '+' : '-',make_ishift(inst_reg&0xfff));
		}
		else{/* imm */
			printf("ldrbt%s\tr%d,[r%d],#%c0x%03x\n",cond_name[cond],rd,rs1,(madd)? '+' : '-',immd);
		}
	}
	else{
		dis_undef();
	}
}

static void dis_pack(void)
{
  int A;
  op1 = (inst_reg>>20)&7;
  op2 = (inst_reg>> 5)&7;
  rd  = (inst_reg>>12)&0xf;
  A   = (inst_reg>>16)&0xf;
  rs1 = NONE;
  rs2 = (inst_reg)&0xf;
  mrot = (inst_reg>>10)&0x3;
  if((op1==2)&&(op2==3)&&(A==0xf)){/*SXTB */
    switch(mrot){
    case 0:
      printf("sxtb%s\tr%d,r%d\n",cond_name[cond],rd,rs2);
      break;
    case 1:
      printf("sxtb%s\tr%d,r%d, ror #8\n",cond_name[cond],rd,rs2);
      break;
    case 2:
      printf("sxtb%s\tr%d,r%d, ror #16\n",cond_name[cond],rd,rs2);
      break;
    case 3:
      printf("sxtb%s\tr%d,r%d, ror #24\n",cond_name[cond],rd,rs2);
      break;
    }
    return;
  }
  if((op1==3)&&(op2==3)&&(A==0xf)){/*SXTH */
    switch(mrot){
    case 0:
      printf("sxth%s\tr%d,r%d\n",cond_name[cond],rd,rs2);
      break;
    case 1:
      printf("sxth%s\tr%d,r%d, ror #8\n",cond_name[cond],rd,rs2);
      break;
    case 2:
      printf("sxth%s\tr%d,r%d, ror #16\n",cond_name[cond],rd,rs2);
      break;
    case 3:
      printf("sxth%s\tr%d,r%d, ror #24\n",cond_name[cond],rd,rs2);
      break;
    }
    return;
  }
  if((op1==6)&&(op2==3)&&(A==0xf)){/*UXTB */
    switch(mrot){
    case 0:
      printf("uxtb%s\tr%d,r%d\n",cond_name[cond],rd,rs2);
      break;
    case 1:
      printf("uxtb%s\tr%d,r%d, ror #8\n",cond_name[cond],rd,rs2);
      break;
    case 2:
      printf("uxtb%s\tr%d,r%d, ror #16\n",cond_name[cond],rd,rs2);
      break;
    case 3:
      printf("uxtb%s\tr%d,r%d, ror #24\n",cond_name[cond],rd,rs2);
      break;
    }
    return;
  }
  if((op1==7)&&(op2==3)&&(A==0xf)){/*UXTH */
    switch(mrot){
    case 0:
      printf("uxth%s\tr%d,r%d\n",cond_name[cond],rd,rs2);
      break;
    case 1:
      printf("uxth%s\tr%d,r%d, ror #8\n",cond_name[cond],rd,rs2);
      break;
    case 2:
      printf("uxth%s\tr%d,r%d, ror #16\n",cond_name[cond],rd,rs2);
      break;
    case 3:
      printf("uxth%s\tr%d,r%d, ror #24\n",cond_name[cond],rd,rs2);
      break;
    }
    return;
  }
  dis_undef();
}

static void dis_media(void)
{
	op  = (inst_reg>>20)&0x1f;
	op2 = (inst_reg>> 5)&0x7;
	rd  = (inst_reg>>16)&0xf;
	rs1 = (inst_reg>> 0)&0xf;
	rs2 = (inst_reg>> 8)&0xf;
	if(op==0x11){
		printf("sdiv%s\tr%d,r%d,r%d\n",cond_name[cond],rd,rs1,rs2);
		return;
	}
	if(op==0x13){
		printf("udiv%s\tr%d,r%d,r%d\n",cond_name[cond],rd,rs1,rs2);
		return;
	}
	if(((op>>3)&3)==1){
	  dis_pack();
	  return;
	}
	if(((op&0x1e)==0x1a)&&((op2&3)==2)){/* SBFX */
	  mlsb = (inst_reg>>7)&0x1f;
	  mwidth=((inst_reg>>16)&0x1f)+1;
	  rd = (inst_reg>>12)&0xf;
	  printf("sbfx%s\tr%d,r%d,#0x%x,#0x%x\n",cond_name[cond],rd,rs1,mlsb,mwidth);
	  return;
	}
	if(((op&0x1e)==0x1c)&&((op2&3)==0)){/* BFC BFI */
	  if(rs1==15){/* BFC */
	    mlsb = (inst_reg>>7)&0x1f;
	    mwidth=((inst_reg>>16)&0x1f)-mlsb+1;
	    rd = (inst_reg>>12)&0xf;
	    printf("bfc%s\tr%d,#0x%x,#0x%x\n",cond_name[cond],rd,mlsb,mwidth);
	  }
	  else{/* BFI */
	    mlsb = (inst_reg>>7)&0x1f;
	    mwidth=((inst_reg>>16)&0x1f)-mlsb+1;
	    rd = (inst_reg>>12)&0xf;
	    printf("bfi%s\tr%d,r%d,#0x%x,#0x%x\n",cond_name[cond],rd,rs1,mlsb,mwidth);
	  }
	  return;
	}
	if(((op&0x1e)==0x1e)&&((op2&3)==2)){/* UBFX */
	  mlsb = (inst_reg>>7)&0x1f;
	  mwidth=((inst_reg>>16)&0x1f)+1;
	  rd = (inst_reg>>12)&0xf;
	  printf("ubfx%s\tr%d,r%d,#0x%x,#0x%x\n",cond_name[cond],rd,rs1,mlsb,mwidth);
	  return;
	}

	dis_undef();
}

static void dis_branch_block()
{
	int R;
	op  = (inst_reg>>20)&0x3f;
	rs1 = (inst_reg>>16)&0xf;
	R   = (inst_reg>>15)&0x1;
	if((op&0x30)==0x20){/* Branch */
		immd = (inst_reg &0xffffff);
		if(immd & (1<<23))
			immd = immd | 0xff000000;
		immd = immd<<2;
		printf("b%s\t0x%08x\n",cond_name[cond],PC+8+immd);
		return;
	}
	if((op&0x30)==0x30){/* Branch with Link */
		immd = (inst_reg &0xffffff);
		if(immd & (1<<23))
			immd = immd | 0xff000000;
		immd = immd<<2;
		printf("bl%s\t0x%08x\n",cond_name[cond],PC+8+immd);
		return;
	}
	rs1 = (inst_reg>>16)&0xf;
	mlist = inst_reg&0xffff;
	
	switch((inst_reg>>23)&0x3){
	case 0:
		amode = "da";
		break;
	case 1:
        	amode = "ia";
		break;
	case 2:
		amode = "db";
		break;
	case 3:
		amode = "ib";
		break;
	}
	mwback = (inst_reg>>21)&1;
	muser  = (inst_reg>>22)&1;
	mexret = (inst_reg>>15)&1;
	mexret = mexret & muser;
	if(op&1)
		printf("ldm%s%s\tr%d%s,%s%s\n",amode,cond_name[cond],rs1,(mwback)? "!" : "",make_rlist(mlist),(mexret)? "^" : "");
	else
	        printf("stm%s%s\tr%d%s,%s\n",amode,cond_name[cond],rs1,(mwback)? "!" : "",make_rlist(mlist));
}

static void dis_copro()
{
	dis_undef();
}

void dis(void)
{
  cond    = (inst_reg>>28)&0xf;
  opcode  = (inst_reg>>25)&0x7;
  opcode2 = (inst_reg>>4)&0x1;

  if(cond==0xf){
    dis_uncond();
    return;
  }

  switch(opcode){
  case 0x0:
  case 0x1:
    dis_data_misc();
    break;
  case 0x2:
    dis_ldst_ub();
    break;
  case 0x3:
    if(opcode2)
      dis_media();
    else
      dis_ldst_ub();
    break;
  case 0x4:
  case 0x5:
    dis_branch_block();
    break;
  default:
    dis_copro();
    break;
  }
}

void disasm(unsigned int addr1, unsigned int addr2)
 {
   PC = (int)addr1;
   while(1){
     inst_reg = get_mem(PC,WORD);
     printf("%08x %08x ",PC, inst_reg);
     dis();
     PC = PC+4;
     if(((unsigned int)PC)>addr2)
       break;
   }
 }
