#include "define.h"
int rs1;
int rs2;
int rs3;
int rd;
int immd;
int OP;
int OP2;
int COND;
int msize;
int msex;

int inst_reg;
int inst_reg2;

static int opcode;


static void simple_dec(void)
{
	opcode = (inst_reg>>9)&0x1f;
	
	switch(opcode){
		case 0x00:
		case 0x01:
		case 0x02:
		case 0x03:/* Logical Shift Left */
			rs1	= (inst_reg>>3)&0x7;
			rs2	= IMMD;
			rd	= inst_reg&0x7;
			immd = (inst_reg>>6)&0x1f;
			OP	= SLL;
			OP2	= FLAG;
			break;
		case 0x04:
		case 0x05:
		case 0x06:
		case 0x07:/* Logical Shift Right */
			rs1	= (inst_reg>>3)&0x7;
			rs2	= IMMD;
			rd	= inst_reg&0x7;
			immd = (inst_reg>>6)&0x1f;
			OP	= SRL;
			OP2	= FLAG;
			break;
		case 0x08:
		case 0x09:
		case 0x0a:
		case 0x0b:/* Arithmetic Shift Right */
			rs1	= (inst_reg>>3)&0x7;
			rs2	= IMMD;
			rd	= inst_reg&0x7;
			immd = (inst_reg>>6)&0x1f;
			OP	= SRA;
			OP2	= FLAG;
			break;
		case 0x0c:/* Add Register */
			rs1	= (inst_reg>>3)&0x7;
			rs2	= (inst_reg>>6)&0x7;
			rd	= inst_reg&0x7;
			immd = 0;
			OP	= ADD;
			OP2	= FLAG;
			break;
		case 0x0d:/* Subtract Register */
			rs1	= (inst_reg>>3)&0x7;
			rs2	= (inst_reg>>6)&0x7;
			rd	= inst_reg&0x7;
			immd = 0;
			OP	= SUB;
			OP2	= FLAG;
			break;
		case 0x0e:/* Add 3bit */
			rs1	= (inst_reg>>3)&0x7;
			rs2	= IMMD;
			rd	= inst_reg&0x7;
			immd = (inst_reg>>6)&0x7;
			OP	= ADD;
			OP2	= FLAG;
			break;
		case 0x0f:/* Subtract 3bit */
			rs1	= (inst_reg>>3)&0x7;
			rs2	= IMMD;
			rd	= inst_reg&0x7;
			immd = (inst_reg>>6)&0x7;
			OP	= SUB;
			OP2	= FLAG;
			break;
		case 0x10:
		case 0x11:
		case 0x12:
		case 0x13:/* Move */
			rs1	= IMMD;
			rs2	= NONE;
			rd	= (inst_reg>>8)&0x7;
			immd    = inst_reg&0xff;
			OP	= MOV;
			OP2	= FLAG;
			break;
		case 0x14:
		case 0x15:
		case 0x16:
		case 0x17:/* Compare */
			rs1 = (inst_reg>>8)&0x7;
			rs2	= IMMD;
			rd      = NONE;
			immd    = inst_reg&0xff;
			OP	= SUB;
			OP2	= FLAG;
			break;
		case 0x18:
		case 0x19:
		case 0x1a:
		case 0x1b:/* Add 8-bit */
			rs1 = (inst_reg>>8)&0x7;
			rs2	= IMMD;
			rd	= (inst_reg>>8)&0x7;
			immd    = inst_reg&0xff;
			OP	= ADD;
			OP2	= FLAG;
			break;
		default:/* Subtract 8-bit */
			rs1 = (inst_reg>>8)&0x7;
			rs2	= IMMD;
			rd	= (inst_reg>>8)&0x7;
			immd    = inst_reg&0xff;
			OP	= SUB;
			OP2	= FLAG;
			break;
	}
}

static void data_dec(void)
{
	opcode = (inst_reg>>6)&0xf;
	
	rs1	= inst_reg&0x7;
	rs2	= (inst_reg>>3)&0x7;
	rd	= inst_reg&0x7;
	immd = 0;
	
	switch(opcode){
		case 0x0:
			OP	= AND;
			OP2	= FLAG;
			break;
		case 0x1:
			OP	= XOR;
			OP2	= FLAG;
			break;
		case 0x2:
			OP	= SLL;
			OP2	= FLAG;
			break;
		case 0x3:
			OP	= SRL;
			OP2	= FLAG;
			break;
		case 0x4:
			OP	= SRA;
			OP2	= FLAG;
			break;
		case 0x5:
			OP	= ADC;
			OP2	= FLAG;
			break;
		case 0x6:
			OP	= SBC;
			OP2	= FLAG;
			break;
		case 0x7:
			OP	= ROR;
			OP2	= FLAG;
			break;
		case 0x8:
			rd	= NONE;
			OP	= AND;
			OP2	= FLAG;
			break;
		case 0x9:
			rs1	= IMMD;
			OP	= SUB;
			OP2	= FLAG;
			break;
		case 0xa:
			rd	= NONE;
			OP	= SUB;
			OP2	= FLAG;
			break;
		case 0xb:
			rd	= NONE;
			OP	= NSUB;
			OP2	= FLAG;
			break;
		case 0xc:
			OP	= OR;
			OP2	= FLAG;
			break;
		case 0xd:
			OP	= MUL;
			OP2	= FLAG;
			break;
	        case 0xe:
			OP	= NAND;
			OP2	= FLAG;
			break;
		default:
			OP	= NOT;
			OP2	= FLAG;
			break;
	}
}

static void special_dec(void)
{
	opcode = (inst_reg>>6)&0xf;
	
	switch(opcode){
		case 0x0:
		case 0x1:
		case 0x2:
		case 0x3:
			rs1	= (((inst_reg>>7)&0x1)<<3)|(inst_reg&0x7);
			rs2	= (inst_reg>>3)&0xf;
			rd	= rs1;
			OP	= ADD;
			OP2	= NOP;
			break;
		case 5:
		case 6:
		case 7:
			rs1	= (((inst_reg>>7)&0x1)<<3)|(inst_reg&0x7);
			rs2	= (inst_reg>>3)&0xf;
			rd	= NONE;
			OP	= SUB;
			OP2	= FLAG;
			break;
		case 8:
		case 9:
		case 0xa:
		case 0xb:
			rs1	= (inst_reg>>3)&0xf;
			rs2	= NONE;
			rd	= (((inst_reg>>7)&0x1)<<3)|(inst_reg&0x7);
			OP	= MOV;
			OP2	= NOP;
			break;
		case 0xc:
		case 0xd:
			rs1	= NONE;
			rs2	= (inst_reg>>3)&0xf;
			rd	= NONE;
			OP	= NOP;
			OP2	= JPR;
			break;
		case 0xe:
		case 0xf:
			rs1	= NEXT_PC;
			rs2	= (inst_reg>>3)&0xf;
			rd	= 14;
			OP	= MOV;
			OP2	= JPR;
			break;
		default:
			OP	= UDF;
			OP2	= NOP;
			break;
	}
}

static void loadpc_dec(void)
{
	rs1	= NEXT3_PC;
	rs2	= IMMD;
	rd	= (inst_reg>>8)&0x7;
	immd = (inst_reg&0xff)<<2;
	OP	= ADD;
	OP2	= LD;
	msize	= WORD;
	msex	= 0;
}

static void load_store_dec(void)
{
	int opA, opB;
	
	opA = (inst_reg>>12)&0xf;
	opB = (inst_reg>> 9)&0x7;
	
	OP	= ADD;
	switch(opA){
		case 5:
			switch(opB){
				case 0:
					OP2	= ST;
					msize	= WORD;
					msex	= 0;
					break;
				case 1:
					OP2	= ST;
					msize	= HWORD;
					msex	= 0;
					break;
				case 2:
					OP2	= ST;
					msize	= BYTE;
					msex	= 0;
					break;
				case 3:
					OP2	= LD;
					msize	= BYTE;
					msex	= 1;
					break;
				case 4:
					OP2	= LD;
					msize	= WORD;
					msex	= 0;
					break;
				case 5:
					OP2	= LD;
					msize	= HWORD;
					msex	= 0;
					break;
				case 6:
					OP2	= LD;
					msize	= BYTE;
					msex	= 0;
					break;
				default:
					OP2	= LD;
					msize	= HWORD;
					msex	= 1;
					break;
			}
			if(OP2==ST){
				rs1	= (inst_reg>>3)&0x7;
				rs2	= (inst_reg>>6)&0x7;
				rs3	= (inst_reg>>0)&0x7;
				rd	= NONE;
			}
			else{
				rs1	= (inst_reg>>3)&0x7;
				rs2	= (inst_reg>>6)&0x7;
				rs3	= NONE;
				rd	= (inst_reg>>0)&0x7;
			}
			break;
		case 6:
			if(opB&0x4){
				OP2	= LD;
				rs1	= (inst_reg>>3)&0x7;
				rs2	= IMMD;
				rs3	= NONE;
				rd	= inst_reg&0x7;
				immd	= ((inst_reg>>6)&0x1f)<<2;
			}
			else{
				OP2	= ST;
				rs1	= (inst_reg>>3)&0x7;
				rs2	= IMMD;
				rs3	= inst_reg&0x7;
				rd	= NONE;
				immd	= ((inst_reg>>6)&0x1f)<<2;
			}
			msize	= WORD;
			msex	= 0;
			break;
		case 7:
			if(opB&0x4){
				OP2	= LD;
				rs1	= (inst_reg>>3)&0x7;
				rs2	= IMMD;
				rs3	= NONE;
				rd	= inst_reg&0x7;
				immd	= (inst_reg>>6)&0x1f;
			}
			else{
				OP2	= ST;
				rs1	= (inst_reg>>3)&0x7;
				rs2	= IMMD;
				rs3	= inst_reg&0x7;
				rd	= NONE;
				immd	= (inst_reg>>6)&0x1f;
			}
			msize	= BYTE;
			msex	= 0;
			break;
		case 8:
			if(opB&0x4){
				OP2	= LD;
				rs1	= (inst_reg>>3)&0x7;
				rs2	= IMMD;
				rs3	= NONE;
				rd	= inst_reg&0x7;
				immd	= ((inst_reg>>6)&0x1f)<<1;
			}
			else{
				OP2	= ST;
				rs1	= (inst_reg>>3)&0x7;
				rs2	= IMMD;
				rs3	= inst_reg&0x7;
				rd	= NONE;
				immd	= ((inst_reg>>6)&0x1f)<<1;
			}
			msize	= HWORD;
			msex	= 0;
			break;
		case 9:
			if(opB&0x4){
				OP2	= LD;
				rs1	= 13;
				rs2	= IMMD;
				rs3	= NONE;
				rd	= (inst_reg>>8)&0x7;
				immd	= (inst_reg&0xff)<<2;
			}
			else{
				OP2	= ST;
				rs1	= 13;
				rs2	= IMMD;
				rs3	= (inst_reg>>8)&0x7;
				rd	= NONE;
				immd	= (inst_reg&0xff)<<2;
			}
			msize	= WORD;
			msex	= 0;
			break;
	}
			
}

static void addr_dec(void)
{
	rs1	= NEXT3_PC;
	rs2	= IMMD;
	rd	= (inst_reg>>8)&0x7;
	immd = (inst_reg&0xff)<<2;
	OP	= ADD;
	OP2	= NOP;
}

static void stack_dec(void)
{
	rs1	= 13;
	rs2	= IMMD;
	rd	= (inst_reg>>8)&0x7;
	immd = (inst_reg&0xff)<<2;
	OP	= ADD;
	OP2	= NOP;
}

static void misc_dec(void)
{
	int opA, opB;
	
	opcode = (inst_reg>>5)&0x7f;
	
	switch(opcode){
		case 0x00:
		case 0x01:
		case 0x02:
		case 0x03:
		case 0x04:
		case 0x05:
		case 0x06:
		case 0x07:
			rs1	= 13;
			rs2	= IMMD;
			rd	= 13;
			immd	= (inst_reg&0x7f)<<2;
			OP	= (opcode&4)? SUB : ADD;
			OP2	= NOP;
			break;
		case 0x10:
		case 0x11:
		case 0x12:
		case 0x13:
			rs1	= (inst_reg>>3)&0x7;
			rs2	= NONE;
			rd	= inst_reg&0x7;
			OP	= SEX;
			OP2	= NOP;
			msize	= (opcode&2)? BYTE : HWORD;
			break;
		case 0x14:
		case 0x15:
		case 0x16:
		case 0x17:
			rs1	= (inst_reg>>3)&0x7;
			rs2	= NONE;
			rd	= inst_reg&0x7;
			OP	= ZEX;
			OP2	= NOP;
			msize	= (opcode&2)? BYTE : HWORD;
			break;
		case 0x20:
		case 0x21:
		case 0x22:
		case 0x23:
		case 0x24:
		case 0x25:
		case 0x26:
		case 0x27:
		case 0x28:
		case 0x29:
		case 0x2a:
		case 0x2b:
		case 0x2c:
		case 0x2d:
		case 0x2e:
		case 0x2f:
			OP	= PUSH;
			OP2	= NOP;
			rd      = NONE;
			immd	= (inst_reg&0xff)|(((inst_reg>>8)&1)<<14);
			break;
		case 0x33:
			OP	= CPS;
			OP2	= NOP;
			immd	= (inst_reg>>4)&1;
			break;
		case 0x50:
		case 0x51:
		case 0x52:
		case 0x53:
		case 0x56:
		case 0x57:
			OP	= REV;
			OP2	= NOP;
			rs1	= (inst_reg>>3)&0x7;
			rs2	= NONE;
			rd	= inst_reg&0x7;
			immd	= (opcode>>1)&0x3;
			break;
		case 0x60:
		case 0x61:
		case 0x62:
		case 0x63:
		case 0x64:
		case 0x65:
		case 0x66:
		case 0x67:
		case 0x68:
		case 0x69:
		case 0x6a:
		case 0x6b:
		case 0x6c:
		case 0x6d:
		case 0x6e:
		case 0x6f:
			OP	= POP;
			OP2	= NOP;
			rd      = NONE;
			immd	= (inst_reg&0xff)|(((inst_reg>>8)&1)<<15);
			break;
		case 0x70:
		case 0x71:
		case 0x72:
		case 0x73:
		case 0x74:
		case 0x75:
		case 0x76:
		case 0x77:
			OP	= BRK;
			OP2	= NOP;
			immd	= inst_reg&0xff;
			break;
		case 0x78:
		case 0x79:
		case 0x7a:
		case 0x7b:
		case 0x7c:
		case 0x7d:
		case 0x7e:
		case 0x7f:
			opA	= (inst_reg>>4)&0xf;
			opB	= inst_reg&0xf;
			if(opB!=0){
				OP	= UDF;
				OP2	= NOP;
			}
			else{
				switch(opA){
					case 0:
						OP	= NOP;
						OP2	= NOP;
						break;
					case 1:
						OP	= YIELD;
						OP2	= NOP;
						break;
					case 2:
						OP	= WFE;
						OP2	= NOP;
						break;
					case 3:
						OP	= WFI;
						OP2	= NOP;
						break;
					case 4:
						OP	= SEV;
						OP2	= NOP;
						break;
					default:
						OP	= UDF;
						OP2	= NOP;
						break;
				}
			}
			break;
		default:
			OP	= UDF;
			OP2	= NOP;
			break;
	}
}

static void stm_dec(void)
{
	OP	= STM;
	OP2	= NOP;
	immd	= inst_reg&0xff;
	rs1	= (inst_reg>>8)&0x7;
	rd      = NONE;
}

static void ldm_dec(void)
{
	OP	= LDM;
	OP2	= NOP;
	immd	= inst_reg&0xff;
	rs1	= (inst_reg>>8)&0x7;
	rd      = NONE;
}

static void jcc_dec(void)
{
	opcode	= (inst_reg>>8)&0xf;
	
	switch(opcode){
		case 0xe:
			OP	= UDF;
			OP2	= NOP;
			rs1	= NONE;
			rs2	= NONE;
			rd	= NONE;
			immd    = inst_reg &0xff;
			break;
		case 0xf:
			OP	= SVC;
			OP2	= NOP;
			rs1	= NONE;
			rs2	= NONE;
			rd	= NONE;
			immd    = inst_reg &0xff;
			break;
		default:
			OP	= ADD;
			OP2	= JCC;
			rs1	= NEXT2_PC;
			rs2	= IMMD;
			rd	= NONE;
			immd	= (inst_reg &0xff);
			immd	= (immd &0x80)? 0xffffff00| immd : immd;
			immd	= immd <<1;
			COND	= (inst_reg>>8)&0xf;
			break;
	}
}

static void jmp_dec(void)
{
	OP	= ADD;
	OP2	= JCC;
	rs1	= NEXT2_PC;
	rs2	= IMMD;
	rd	= NONE;
	immd	= (inst_reg &0x7ff);
	immd	= (immd &0x400)? 0xfffff800| immd : immd;
	immd	= immd <<1;
	COND	= 0xe;/* always */
}

static void long_dec(void)
{
	int op1,op2;
	
	op1	= (inst_reg>>11)&3;
	op2	= (inst_reg2>>15)&1;
	if((op1==2)&&(op2==1)){
		op1	= (inst_reg>>4)&0x7f;
		op2	= (inst_reg2>>12)&0x7;
		if(((op2&0x5)==0)&&((op1&0x7e)==0x38)){
			OP	= MSR;
			rs1	= inst_reg&0xf;
			rs2	= NONE;
			rd	= inst_reg2&0x1f;
		}
		else if(((op2&0x5)==0)&&((op1&0x7f)==0x3b)){
			op1	= (inst_reg2>>4)&0xf;
			if(op1==4){
				OP	= DSB;
				OP2	= NOP;
				immd	= inst_reg2&0xf;
			}
			else if(op1==5){
				OP	= DMB;
				OP2	= NOP;
				immd	= inst_reg2&0xf;
			}
			else if(op1==6){
				OP	= ISB;
				OP2	= NOP;
				immd	= inst_reg2&0xf;
			}
			else{
				OP	= UDF;
				OP2	= NOP;
			}
		}
		else if(((op2&0x5)==0)&&((op1&0x7f)==0x3e)){
			OP	= MRS;
			rs1	= (inst_reg2>>0)&0x1f;
			rs2	= NONE;
			rd	= (inst_reg2>>8)&0xf;
		}
		else if((op2==2)&&(op1==0x7f)){
			OP	= UDF;
			OP2	= NOP;
		}
		else if((op2&0x5)==5){
			int s, J1, J2;
			int I1, I2;
			int imm10, imm11;
			s	= (inst_reg>>10)&1;
			J1	= (inst_reg2>>13)&1;
			J2	= (inst_reg2>>11)&1;
			I1	= (J1 ^ s)^1;
			I2	= (J2 ^ s)^1;
			imm10	= inst_reg&0x3ff;
			imm11	= inst_reg2&0x7ff;
			immd	= (s<<23)|(I1<<22)|(I2<<21)|(imm10<<11)|imm11;
			immd	= (immd & (1<<23))? 0xff000000|immd : immd;
			immd	= immd <<1;			
			OP	= ADD;
			OP2	= JAL;
			rs1	= NEXT2_PC;
			rs2	= IMMD;
			rs3	= NEXT_PC;
			rd	= 14;
		}
		else{
			OP	= UDF;
			OP2	= NOP;
		}
	}
	else{
		OP	= UDF;
		OP2	= NOP;
		rs1	= NONE;
		rs2	= NONE;
		rd	= NONE;
	}
}

void decode(void)
{
	opcode = (inst_reg>>10)&0x3f;
	
	switch(opcode){
		case 0x00:
		case 0x01:
		case 0x02:
		case 0x03:
		case 0x04:
		case 0x05:
		case 0x06:
		case 0x07:
		case 0x08:
		case 0x09:
		case 0x0a:
		case 0x0b:
		case 0x0c:
		case 0x0d:
		case 0x0e:
		case 0x0f:
			simple_dec();
			break;
		case 0x10:
			data_dec();
			break;
		case 0x11:
			special_dec();
			break;
	        case 0x12:
		case 0x13:
			loadpc_dec();
			break;
		case 0x14:
		case 0x15:
		case 0x16:
	        case 0x17:
		case 0x18:
		case 0x19:
		case 0x1a:
		case 0x1b:
		case 0x1c:
		case 0x1d:
		case 0x1e:
		case 0x1f:
		case 0x20:
		case 0x21:
		case 0x22:
		case 0x23:
		case 0x24:
		case 0x25:
		case 0x26:
		case 0x27:
			load_store_dec();
			break;
		case 0x28:
		case 0x29:
			addr_dec();
			break;
		case 0x2a:
		case 0x2b:
			stack_dec();
			break;
		case 0x2c:
		case 0x2d:
		case 0x2e:
		case 0x2f:
			misc_dec();
			break;
		case 0x30:
		case 0x31:
			stm_dec();
			break;
		case 0x32:
		case 0x33:
			ldm_dec();
			break;
		case 0x34:
		case 0x35:
		case 0x36:
		case 0x37:
			jcc_dec();
			break;
		case 0x38:
		case 0x39:
			jmp_dec();
			break;
		default:
			long_dec();
			break;
	}
}
		
		
