#include "define.h"

extern int PC;
extern int immd;
extern int rs1;
extern int rs2;
extern int rs3;
extern int sa;
extern int rd;
extern int funct3;
extern int funct7;
extern int OP;
extern int OP2;
extern int msize;
extern int msex;
extern int cond;

static int target;
static int in_delay_slot=0;

int get_value(int reg)
{
  if(reg == 0)
    return 0;
  if((reg>=1)&&(reg<=31))
    return get_reg(reg);
  if(reg == 32)
    return immd;
  if(reg == 33)
    return PC+8;
  if(reg == 34)
    return PC+4;
  if(reg == 35)
    return  get_reg(reg);
  if(reg == 36)
    return  get_reg(reg);
  return 0;
}

int compare(int src1, int src2)
{
  unsigned int us1, us2;
	
  us1	= (unsigned int)src1;
  us2	= (unsigned int)src2;
	
  switch(cond){
  case EQ:
    return (src1 == src2);
  case NE:
    return (src1 != src2);
  case LT:
    return (src1 < src2);
  case GE:
    return (src1 >= src2);
  case LTU:
    return (us1 < us2);
  case GEU:
    return (us1 >= us2);
  case LE:
    return (src1 <= src2);
  case GT:
    return (src1 > src2);
  default:
    return 0;
  }
}

int operation(int src1, int src2)
{
  unsigned long long s1,s2,d;	
  unsigned int us1,us2;
  int d_tmp;
	
  us1	= (unsigned int)src1;
  us2	= (unsigned int)src2;
	
  switch(OP){
  case ADD:
    d_tmp = src1 + src2;
    if(
       ((src1>=0)&&(src2>=0)&&(d_tmp<0))
       ||((src1<0)&&(src2<0)&&(d_tmp>=0))
       ){
      OP2 = EXP;
      rd  = 0;
    }
    return d_tmp;
  case ADDU:
    return src1 + src2;
  case SUB:
    d_tmp = src1 - src2;
    if(
       ((src1>=0)&&(src2<0)&&(d_tmp<0))
       ||((src1<0)&&(src2>=0)&&(d_tmp>=0))
       ){
      OP2 = EXP;
      rd  = 0;
    }
    return d_tmp;
  case SUBU:
    return  src1 - src2;
  case MUL:
    s1	= (unsigned long long)((src1<0)? 0xffffffff00000000LL | us1 : us1);
    s2	= (unsigned long long)((src2<0)? 0xffffffff00000000LL | us2 : us2);
    d	= s1 * s2;
    put_reg(35,(int)((d>>32)&0xffffffff));
    put_reg(36,(int)((d>> 0)&0xffffffff));
    rd = 0;
    return 0;
  case MULU:
    s1	= (unsigned long long)us1;
    s2	= (unsigned long long)us2;
    d = s1 * s2;
    put_reg(35,(int)((d>>32)&0xffffffff));
    put_reg(36,(int)((d>> 0)&0xffffffff));
    rd = 0;
    return 0;
  case DIV:
    put_reg(35,src1 % src2);
    put_reg(36,src1 / src2);
    rd = 0;
    return 0;
  case DIVU:
    put_reg(35,us1 % us2);
    put_reg(36,us1 / us2);
    rd = 0;
    return 0;
  case AND:
    return src1 & src2;
  case OR:
    return src1 | src2;
  case NOR:
    return ~(src1 | src2);
  case XOR:
    return  src1 ^ src2;
  case SLL:
    return src1 << sa;
  case SRA:
    return src1 >> sa;
  case SRL:
    return us1 >> sa;
  case SLLV:
    src2 &= 0x1f;
    return src1 << src2;
  case SRAV:
    src2 &= 0x1f;
    return src1 >> src2;
  case SRLV:
    src2 &= 0x1f;
    return us1 << src2;
  }
  return 0;
}

void execute(void)
{
  int word, word2;
  int v_src1, v_src2, v_src3, v_dst;

  v_src1	= get_value(rs1);
  v_src2	= get_value(rs2);
  v_src3	= get_value(rs3);

  switch(OP){
  case MOV:
    v_dst	= v_src2;
    break;
  case CMP:/* for condition branch and SET */
    v_dst	= compare(v_src1, v_src2);
    break;	
  default:
    v_dst	= operation(v_src1, v_src2);
    break;
  }
  
  put_reg(rd, v_dst);

  switch(OP2){
  case JMP:
    target = ((PC+4)&0xf0000000)|(immd<<2);
    in_delay_slot = 1;
    break;		
  case JPR:
    target = v_src1;
    in_delay_slot = 1;
    break;
  case JCC:
    if(v_dst){
      target	= PC + 4 + immd;
      in_delay_slot = 1;
    }
    if(rd!=0)
      put_reg(31,PC+8);
    break;
  case LD:
    switch(msize){
    case BYTE:
      word = get_mem(v_dst, BYTE);
      if(msex){
	if(word & 0x80)
	  word |= 0xffffff00;
	else
	  word &= 0x000000ff;
      }
      else{
	  word &= 0x000000ff;
      }
      break;
    case HWORD:
      word = get_mem(v_dst, HWORD);
      if(msex){
	if(word & 0x8000)
	  word |= 0xffff0000;
	else
	  word &= 0x0000ffff;
      }
      else{
	  word &= 0x0000ffff;
      }
      break;
    case WORD:
      word = get_mem(v_dst, WORD);
      break;
    case LWL:
      word = get_mem(v_dst, WORD);
      word2 = get_reg(rd);
      switch(v_dst&3){
      case 0:
	word = (word<<24)|(word2&0x00ffffff);
	break;
      case 1:
	word = (word<<16)|(word2&0x0000ffff);
	break;
      case 2:
	word = (word<< 8)|(word2&0x000000ff);
      case 3:
	word = word;
	break;
      }
      break;
    case LWR:
      word = get_mem(v_dst, WORD);
      word2 = get_reg(rd);
      switch(v_dst&3){
      case 0:
	word = word;
	break;
      case 1:
	word = (word2&0xff000000)|((word>> 8)&0x00ffffff);
	break;
      case 2:
	word = (word2&0xffff0000)|((word>>16)&0x0000ffff);
      case 3:
	word = (word2&0xffffff00)|((word>>24)&0x000000ff);
	break;
      }
      break;
    }
    put_reg(rd, word);
    break;		
  case ST:
    switch(msize){
    case BYTE:
      put_mem(v_dst, v_src3, BYTE);
      break;
    case HWORD:
      put_mem(v_dst, v_src3, HWORD);
      break;
    case WORD:
      put_mem(v_dst, v_src3, WORD);
      break;
    }
    break;
  case SWL:
    word = get_mem(v_dst, WORD);
    switch(v_dst&3){
    case 0:
      word = (word&0xffffff00)|((v_src3>>24)&0x000000ff);
      break;
    case 1:
      word = (word&0xffff00000)|((v_src3>>16)&0x0000ffff);
      break;
    case 2:
      word = (word&0xff0000000)|((v_src3>> 8)&0x00ffffff);
      break;
    case 3:
      word = v_src3;
      break;
    }
    put_mem(v_dst, word, WORD);
    break;
  case SWR:
    word = get_mem(v_dst, WORD);
    switch(v_dst&3){
    case 0:
      word = v_src3;
      break;
    case 1:
      word = (word&0x000000ff)|((v_src3<< 8)&0xff000000);
      break;
    case 2:
      word = (word&0x0000ffff)|((v_src3<<16)&0xffff0000);
      break;
    case 3:
      word = (word&0x00ffffff)|((v_src3<<24)&0xff000000);
      break;
    }
    put_mem(v_dst, word, WORD);
    break;
  default:
    break;
  }
  if(in_delay_slot==1){
    in_delay_slot = 2;
    PC = PC + 4;
  }
  else if (in_delay_slot==2){
    in_delay_slot = 0;
    PC = target;
  }
  else {
    in_delay_slot = 0;
    PC = PC + 4;
  }

}
