#include "define.h"

extern int PC;
extern int immd;
extern int rs1;
extern int rs2;
extern int rs3;
extern int rd;
extern int funct3;
extern int funct7;
extern int OP;
extern int OP2;

int get_value(int reg)
{
  if(reg == 0)
    return 0;
  if((reg>=1)&&(reg<=31))
    return get_reg(reg);
  if(reg == 32)
    return immd;
  if(reg == 33)
    return PC;
  if(reg == 34)
    return PC+4;
  return 0;
}

int compare(int src1, int src2)
{
  unsigned int us1, us2;
	
  us1	= (unsigned int)src1;
  us2	= (unsigned int)src2;
	
  switch(funct3){
  case 0:/* EQ */
    return (src1 == src2);
  case 1:/* NE */
    return (src1 != src2);
  case 4:/* LT */
    return (src1 < src2);
  case 5:/* GE */
    return (src1 >= src2);
  case 6:/* LTU */
    return (us1 < us2);
  case 7:/* GEU */
    return (us1 >= us2);
  default:
    return 0;
  }
}

int operation(int src1, int src2)
{
  unsigned long long s1,s2,d;	
  unsigned int us1,us2;
	
  us1	= (unsigned int)src1;
  us2	= (unsigned int)src2;
	
  if(funct7==1){/* RV32M */
    switch(funct3){
    case 0:/* MUL */
      return (src1 * src2);
    case 1:/* MULH */
      s1	= (unsigned long long)((src1<0)? 0xffffffff00000000LL | us1 : us1);
      s2	= (unsigned long long)((src2<0)? 0xffffffff00000000LL | us2 : us2);
      d	= s1 * s2;
      return (d>>32)&0xffffffff;
    case 2:/* MULHSU */
      s1	= (unsigned long long)((src1<0)? 0xffffffff00000000LL | us1 : us1);
      s2	= (unsigned long long)us2;
      d	= s1 * s2;
      return (d>>32)&0xffffffff;
    case 3:/* MULHU */
      s1	= (unsigned long long)us1;
      s2	= (unsigned long long)us2;
      d	= s1 * s2;
      return (d>>32)&0xffffffff;
    case 4:/* DIV */
      return (src1 / src2);
    case 5:/* DIVU */
      return (us1 / us2);
    case 6:/* REM */
      return (src1 % src2);
    case 7:/* REMU */
      return (us1 % us2);
    default:
      return 0;
    }		
  }

  switch(funct3){
  case 0:/* ADD/SUB */
    if(funct7)
      return (src1 - src2);
    else
      return (src1 + src2);
  case 1:/* SL */
    return (src1 << src2);
  case 2:/* SLT */
    return (src1 < src2);
  case 3:/* SLTU */
    return (us1 < us2);
  case 4:/* XOR */
    return (src1 ^ src2);
  case 5:/* SR */
    if(funct7)
      return (src1 >> src2);
    else
      return (us1 >> us2);
  case 6:/* OR */
    return (src1 | src2);
  case 7:/* AND */
    return (src1 & src2);
  }
}

void execute(void)
{
  int word;
  int v_src1, v_src2, v_src3, v_dst;

  v_src1	= get_value(rs1);
  v_src2	= get_value(rs2);
  v_src3	= get_value(rs3);
	
  switch(OP){
  case MOV:
    v_dst	= v_src2;
    break;
  case CMP:/* for condition branch */
    v_dst	= compare(v_src1, v_src2);
    break;	
  case ADDI:
    v_dst	= v_src1 + v_src2;
    break;	
  case ALU:
    if((funct3==1)||(funct3==5))/* shift */
      v_src2 &= 0x1f;
    v_dst	= operation(v_src1, v_src2);
  }
  
  put_reg(rd, v_dst);

  switch(OP2){
  case JMP:
    PC	= PC + immd;
    break;		
  case JPR:
    PC	= v_src1;
    break;
  case JCC:
    if(v_dst)
      PC	= PC + immd;
    else
      PC	= PC + 4;
    break;
  case LD:
    switch(funct3){
    case 0:/* BYTE */
      word = get_mem(v_dst, BYTE);
      if(word & 0x80)
	word |= 0xffffff00;
      else
	word &= 0x000000ff;
      break;
    case 1:/* HWORD */
      word = get_mem(v_dst, HWORD);
      if(word & 0x8000)
	word |= 0xffff0000;
      else
	word &= 0x0000ffff;
      break;
    case 2:/* WORD */
      word = get_mem(v_dst, WORD);
      break;
    case 4:/* BYTE UNSIGNED */
      word = get_mem(v_dst, BYTE);
      word &= 0x000000ff;
      break;
    case 5:/* HWORD UNSIGNED */
      word = get_mem(v_dst, HWORD);
      word &= 0x0000ffff;
      break;
    }
    put_reg(rd, word);
    PC	= PC + 4;
    break;		
  case ST:
    switch(funct3){
    case 0:/* BYTE */
      put_mem(v_dst, v_src3, BYTE);
      break;
    case 1:/* HWORD */
      put_mem(v_dst, v_src3, HWORD);
      break;
    case 2:/* WORD */
      put_mem(v_dst, v_src3, WORD);
      break;
    }
    PC	= PC + 4;
    break;
  default:
    PC	= PC + 4;
  }
}
