#include "define.h"

extern int PC;
extern int immd;
extern int rs1;
extern int rs2;
extern int rs3;
extern int sa;
extern int rd;
extern int funct3;
extern int funct7;
extern int OP;
extern int OP2;
extern int msize;
extern int msex;
extern int mtype;
extern int cond;
extern int delay_slot;

static int target;
static int in_delay_slot=0;

int get_value(int reg)
{
  if((reg>=0)&&(reg<=15))
    return get_reg(reg);
  if(reg==IMMD)
    return immd;
  if(reg==PC_REG){
    if(msize==WORD)
      return get_reg(PC_REG)&0xfffffffc;
    else
      return get_reg(PC_REG);
  }
  if(reg==MACH_REG)
    return get_reg(MACH_REG);
  if(reg==MACL_REG)
    return get_reg(MACL_REG);
  if(reg==GBR_REG)
    return get_reg(GBR_REG);
  if(reg==PR_REG)
    return get_reg(PR_REG);
  if(reg==T_REG)
    return get_reg(T_REG);
  if(reg==SR_REG)
    return get_reg(SR_REG);
  if(reg==VBR_REG)
    return get_reg(VBR_REG);
  return 0;
}

int compare(int src1, int src2)
{
  unsigned int us1, us2;
  int tmp;
	
  us1	= (unsigned int)src1;
  us2	= (unsigned int)src2;
	
  switch(cond){
  case EQ:
    return (src1 == src2);
  case NE:
    return (src1 != src2);
  case LT:
    return (src1 < src2);
  case GE:
    return (src1 >= src2);
  case LTU:
    return (us1 < us2);
  case GEU:
    return (us1 >= us2);
  case LE:
    return (src1 <= src2);
  case GT:
    return (src1 > src2);
  case GTU:
    return (us1 > us2);
  case ALWAYS:
    return 1;
  case TRUE:
    return get_reg(T_REG)==1;
  case FALSE:
    return get_reg(T_REG)==0;
  case DT:
    put_reg(rs1,src1-1);
    return get_reg(rs1)==0;
  case STR:
    tmp = src1 ^ src2;
    return ((tmp & 0xff000000)==0)
         ||((tmp & 0x00ff0000)==0)
         ||((tmp & 0x0000ff00)==0)
         ||((tmp & 0x000000ff)==0);
  case TST:
    tmp = src1 & src2;
    return (tmp == 0);
  default:
    return 0;
  }
}

void div1(unsigned int n, unsigned int m)
{
  unsigned int tmp0;
  unsigned char old_q, tmp1;
  unsigned char T, Q, M;

  T=get_reg(T_REG);
  Q=get_reg(Q_REG);
  M=get_reg(M_REG);
  old_q=Q;
  Q=(unsigned char)((0x80000000 & n)!=0);
  n<<=1;
  n|=(unsigned long)T;
  switch(old_q){
  case 0:
    switch(M){
    case 0:
      tmp0=n;
      n-=m;
      tmp1=(n>tmp0);
      switch(Q){
      case 0:
	Q=tmp1;
	break;
      case 1:
	Q=(unsigned char)(tmp1==0);
	break;
      }
      break;
    case 1:
      tmp0=n;
      n+=m;
      tmp1=(n<tmp0);
      switch(Q){
      case 0:
	Q=(unsigned char)(tmp1==0);
	break;
      case 1:
	Q=tmp1;
	break;
      }
      break;
    }
    break;
  case 1:
    switch(M){
    case 0:
      tmp0=n;
      n+=m;
      tmp1=(n<tmp0);
      switch(Q){
      case 0:
	Q=tmp1;
	break;
      case 1:
	Q=(unsigned char)(tmp1==0);
	break;
      }
      break;
    case 1:
      tmp0=n;
      n-=m;
      tmp1=(n>tmp0);
      switch(Q){
      case 0:
	Q=(unsigned char)(tmp1==0);
	break;
      case 1:
	Q=tmp1;
	break;
      }
      break;
    }
    break;
  }
  T=(Q==M);
  put_reg(T_REG,T);
  put_reg(Q_REG,Q);
  put_reg(M_REG,M);
  put_reg(rs1,n);
  put_reg(rs2,m);
}

void mac(int src1, int src2)
{
  long long s1,s2,a,d,m;
  long long h,l;
  h = (long long)get_reg(MACH_REG) & 0xffffffffLL;
  l = (long long)get_reg(MACL_REG) & 0xffffffffLL;
  a = (h<<32)|l;
  switch(msize){
  case HWORD:
    s1 = (long long)get_mem(src1,HWORD);
    s2 = (long long)get_mem(src2,HWORD);
    if(s1 &0x8000LL)
      s1 |= 0xffffffffffff0000LL;
    else
      s1 &= 0x000000000000ffffLL;
    if(s2 &0x8000LL)
      s2 |= 0xffffffffffff0000LL;
    else
      s2 &= 0x000000000000ffffLL;
    src1 += 2;
    src2 += 2;
    break;
  case WORD:
    s1 = (long long)get_mem(src1,WORD);
    s2 = (long long)get_mem(src2,WORD);
    if(s1 &0x8000LL)
      s1 |= 0xffffffff00000000LL;
    else
      s1 &= 0x00000000ffffffffLL;
    if(s2 &0x8000LL)
      s2 |= 0xffffffff00000000LL;
    else
      s2 &= 0x00000000ffffffffLL;
    src1 += 4;
    src2 += 4;
    break;
  }
  d = s1 * s2 * a;
  put_reg(MACH_REG, (int)((d>>32)&0xffffffffLL));
  put_reg(MACL_REG, (int)((d>> 0)&0xffffffffLL));
  put_reg(rs1, src1);
  put_reg(rs2, src2);
}

int operation(int src1, int src2)
{
  unsigned long long s1,s2,d;	
  unsigned int us1,us2;
  int d_tmp;
  int cf,vf;
	
  us1	= (unsigned int)src1;
  us2	= (unsigned int)src2;

  cf    = 0;
  vf    = 0;
	
  switch(OP){
  case ADD:
    return  src1 + src2;
  case ADDC:
    s1 = (unsigned long long)(src1&0x00000000ffffffffLL);
    s2 = (unsigned long long)(src2&0x00000000ffffffffLL);
    d  = s1 + s2 + (unsigned long long)get_reg(T_REG);
    d_tmp = (int)(d & 0xffffffffLL);
    cf = (int)((d>>32)&1LL);
    put_reg(T_REG,cf);
    return d_tmp;
  case ADDV:
    d_tmp = src1 + src2;
    vf = ((src1>=0)&&(src2>=0)&&(d_tmp<0))
      ||((src1<0)&&(src2<0)&&(d_tmp>=0));
    put_reg(T_REG,vf);
    return d_tmp;
  case AND:
    return src1 & src2;
  case CLRMAC:
    put_reg(MACH_REG,0);
    put_reg(MACL_REG,0);
    return 0;
  case CMP:
    return compare(src1, src2);
  case DIV0S:
    put_reg(Q_REG,(src1>>31)&1);
    put_reg(M_REG,(src2>>31)&1);
    put_reg(T_REG,((src1^src2)>>31)&1);
    return 0;
  case DIV0U:
    put_reg(Q_REG,0);
    put_reg(M_REG,0);
    put_reg(T_REG,0);
    return 0;
  case DIV1:
    div1(src1, src2);
    return 0;
  case MAC:
    mac(src1, src2);
    return 0;
  case MOV:
    return src2;
  case MUL:
    s1 = (unsigned long long)(src1&0x00000000ffffffffLL);
    s2 = (unsigned long long)(src2&0x00000000ffffffffLL);
    if(s1 & 0x80000000LL)
      s1 |= 0xffffffff00000000LL;
    if(s2 & 0x80000000LL)
      s2 |= 0xffffffff00000000LL;
    d = s1 * s2;
    put_reg(MACH_REG,(int)((d>>32)&0xffffffffLL));
    put_reg(MACL_REG,(int)((d>> 0)&0xffffffffLL));
    return 0;
  case MULU:
    s1 = (unsigned long long)(src1&0x00000000ffffffffLL);
    s2 = (unsigned long long)(src2&0x00000000ffffffffLL);
    d = s1 * s2;
    put_reg(MACH_REG,(int)((d>>32)&0xffffffffLL));
    put_reg(MACL_REG,(int)((d>> 0)&0xffffffffLL));
    return 0;
  case MULUW:
    s1 = (unsigned long long)(src1&0x000000000000ffffLL);
    s2 = (unsigned long long)(src2&0x000000000000ffffLL);
    d = s1 * s2;
    put_reg(MACH_REG,(int)((d>>32)&0xffffffffLL));
    put_reg(MACL_REG,(int)((d>> 0)&0xffffffffLL));
    return 0;
  case MULW:
    s1 = (unsigned long long)(src1&0x000000000000ffffLL);
    s2 = (unsigned long long)(src2&0x000000000000ffffLL);
    if(s1 & 0x8000LL)
      s1 |= 0xffffffffffff0000LL;
    if(s2 & 0x8000LL)
      s2 |= 0xffffffffffff0000LL;
    d = s1 * s2;
    put_reg(MACH_REG,(int)((d>>32)&0xffffffffLL));
    put_reg(MACL_REG,(int)((d>> 0)&0xffffffffLL));
    return 0;
  case MULL:
    return src1 * src2;
  case NEG:
    return -src2;
  case NEGC:
    s2 = (unsigned long long)((~src2)&0x00000000ffffffffLL);
    s2 += 1LL-(unsigned long long)get_reg(T_REG);
    cf = (int)((d>>32)&1LL);
    cf ^= 1; /* carry -> borrow */
    put_reg(T_REG,cf);
    return (int)(d&0xffffffffLL);
  case NOT:
    return ~src2;
  case OR:
    return src1 | src2;
  case ROL:
    cf = (src1>>31)&1;
    put_reg(T_REG,cf);
    return (src1<<1)|cf;
  case ROLC:
    cf = get_reg(T_REG);
    put_reg(T_REG,(src1>>31)&1);
    return (src1<<1)|cf;
  case ROR:
    cf = src1&1;
    put_reg(T_REG,cf);
    return (cf<<31)|((src1>>1)&0x7fffffff);
  case RORC:
    cf = get_reg(T_REG);
    put_reg(T_REG, src1 & 1);
    return (cf<<31)|((src1>>1)&0x7fffffff);
  case SEX:
    if(msize==BYTE)
      return (src2&0x80)? 0xffffff00|src2 : src2&0xff;
    else
      return (src2&0x8000)? 0xffff0000|src2 : src2&0xffff;
  case SLA:
  case SLL:
    if(us2==1)
      put_reg(T_REG,(src1>>31)&1);
    return src1 << us2;
  case SRA:
    if(us2==1)
      put_reg(T_REG,src1&1);
    return src1 >> us2;
  case SRL:
    if(us2==1)
      put_reg(T_REG,src1&1);
    return us1 >> us2;
  case SUB:
    return src1 - src2;
  case SUBC:
    s1 = (unsigned long long)(src1&0x00000000ffffffffLL);
    s2 = (unsigned long long)((~src2)&0x00000000ffffffffLL);
    d  = s1 + s2 + 1LL - (unsigned long long)get_reg(T_REG);
    d_tmp = (int)(d & 0xffffffffLL);
    cf = (int)((d>>32)&1LL);
    put_reg(T_REG,cf^1);
    return d_tmp;
  case SUBV:
    d_tmp = src1 - src2;
    vf = ((src1>=0)&&(src2<0)&&(d_tmp<0))
      ||((src1<0)&&(src2>=0)&&(d_tmp>=0));
    put_reg(T_REG,vf);
    return d_tmp;
  case SWAPB:
    d_tmp = src2;
    return (d_tmp&0xffff0000)|((d_tmp&0xff)<<8)|((d_tmp>>8)&0xff);
  case SWAPW:
    d_tmp = src2;
    return ((d_tmp&0xffff)<<16)|((d_tmp>>16)&0xffff);
  case TAS:
    d_tmp = get_mem(src1,BYTE);
    put_reg(T_REG,(d_tmp==0));
    d_tmp |= 0x80;
    put_mem(src1,d_tmp,BYTE);
    return 0;
  case XOR:
    return src1 ^ src2;
  case XTRCT:
    d = ((unsigned long long)us2)<<32LL | (unsigned long long)us1;
    return (int)((d>>16)&0xffffffffLL);
  case ZEX:
    if(msize==BYTE)
      return src2 & 0xff;
    else
      return src2 & 0xffff;
  }
  return 0;
}

void execute(void)
{
  int word, word2;
  int v_src1, v_src2, v_src3, v_dst;

  v_src1	= get_value(rs1);
  v_src2	= get_value(rs2);
  v_src3	= get_value(rs3);

  v_dst	= operation(v_src1, v_src2);

  put_reg(rd, v_dst);

  switch(OP2){
  case BRR:
    target = v_src1 + get_reg(PC_REG);
    in_delay_slot = delay_slot;
    break;
  case JCC:
    if(v_dst){
      target	= get_reg(PC_REG) + immd;
      in_delay_slot = delay_slot;
    }
    break;
  case JPR:
    target = v_src1;
    in_delay_slot = delay_slot;
    break;
  case JSR:
    target = v_dst;
    put_reg(PR_REG,get_reg(PC_REG));
    in_delay_slot = delay_slot;
    break;
  case LD:
    if(mtype==PREDEC){
      switch(msize){
      case BYTE:
	v_dst -= 1;
	break;
      case HWORD:
	v_dst -= 2;
	break;
      case WORD:
	v_dst -=4;
	break;
      }
      put_reg(rs1,v_dst);
    }
    switch(msize){
    case BYTE:
      word = get_mem(v_dst, BYTE);
      if(msex){
	if(word & 0x80)
	  word |= 0xffffff00;
	else
	  word &= 0x000000ff;
      }
      else{
	  word &= 0x000000ff;
      }
      break;
    case HWORD:
      word = get_mem(v_dst, HWORD);
      if(msex){
	if(word & 0x8000)
	  word |= 0xffff0000;
	else
	  word &= 0x0000ffff;
      }
      else{
	  word &= 0x0000ffff;
      }
      break;
    case WORD:
      word = get_mem(v_dst, WORD);
      break;
    }
    put_reg(rd, word);
    if(mtype==POSTINC){
      switch(msize){
      case BYTE:
	v_dst += 1;
	break;
      case HWORD:
	v_dst += 2;
	break;
      case WORD:
	v_dst +=4;
	break;
      }
      put_reg(rs1,v_dst);
    }
    break;
  case ST:
    if(mtype==PREDEC){
      switch(msize){
      case BYTE:
	v_dst -= 1;
	break;
      case HWORD:
	v_dst -= 2;
	break;
      case WORD:
	v_dst -=4;
	break;
      }
      put_reg(rs1,v_dst);
    }
    switch(msize){
    case BYTE:
      put_mem(v_dst, v_src3, BYTE);
      break;
    case HWORD:
      put_mem(v_dst, v_src3, HWORD);
      break;
    case WORD:
      put_mem(v_dst, v_src3, WORD);
      break;
    }
    if(mtype==POSTINC){
      switch(msize){
      case BYTE:
	v_dst += 1;
	break;
      case HWORD:
	v_dst += 2;
	break;
      case WORD:
	v_dst +=4;
	break;
      }
      put_reg(rs1,v_dst);
    }
    break;
  case LDST_AND:
  case LDST_OR:
  case LDST_TST:
  case LDST_XOR:
  case TRAP:
  case SLEEP:
  case RTE:
  case EXP:
  default:
    break;
  }

  if(in_delay_slot==1){
    in_delay_slot = 2;
    put_reg(PC_REG,get_reg(PC_REG)-2);
  }
  else if (in_delay_slot==2){
    in_delay_slot = 0;
    put_reg(PC_REG,target);
  }
  else if(OP2==JCC){
    in_delay_slot = 0;
    if(v_dst)
      put_reg(PC_REG,target);
    else
      put_reg(PC_REG,get_reg(PC_REG)-2);
  }
  else {
    in_delay_slot = 0;
    put_reg(PC_REG,get_reg(PC_REG)-2);
  }
}
