#include <stdio.h>
#include <string.h>
#include "define.h"

static FILE *fp;

static int hex(int c)
{
  if((c >= '0') && (c <= '9')) return ( c - '0');
  if((c >= 'A') && (c <= 'F')) return ((c - 'A') + 10);
  if((c >= 'a') && (c <= 'f')) return ((c - 'a') + 10);
  return 0;
}

static void put_byte(long addr, int byte)
{
  int word;

  byte &= 0xff;
  word = get_word(addr);
  switch(addr&3){
  case 0:
    word = (word&0xffffff00)|byte;
    break;
  case 1:
    word = (word&0xffff00ff)|(byte<<8);
    break;
  case 2:
    word = (word&0xff00ffff)|(byte<<16);
    break;
  case 3:
    word = (word&0x00ffffff)|(byte<<24);
    break;
  }
  put_word(addr,word);
}

/*
  Intel Hex Load
*/
void loadi(char *fname, char *base)
{

  int eof, i;
  int ch;

  int	reclen, adofset, rectyp;
  long  adbase, address;
  int	byte;

  adbase = 0;
  fp = fopen(fname,"r");
  if(fp==NULL){
    fprintf(stderr,"cannot open %s\n",fname);
    return;
  }
  eof = 0;
  while(!eof){
    while((ch=getc(fp)) != ':')
      if(ch==EOF){
	fclose(fp);
	return;
      }
    reclen = hex( getc(fp) );
    reclen = reclen * 16 + hex( getc(fp) );
    adofset = hex( getc(fp) );
    adofset = adofset * 16 + hex( getc(fp) );
    adofset = adofset * 16 + hex( getc(fp) );
    adofset = adofset * 16 + hex( getc(fp) );
    rectyp = hex( getc(fp) );
    rectyp = rectyp * 16 + hex( getc(fp) );
    switch( rectyp ){
    case 0x00: /* data record */
      address = adbase + adofset;
      for(i=0; i<reclen ; i++){
	byte = hex( getc(fp) );
	byte = byte * 16 + hex( getc(fp) );
	put_byte(address+i, byte);
      }
      break;
    case 0x01:  /* end of file */
      fclose( fp );
      eof = 1;
      break;
    case 0x02:  /* extended address code */
      address = hex( getc(fp) );
      address = address * 16 + hex( getc(fp) );
      address = address * 16 + hex( getc(fp) );
      address = address * 16 + hex( getc(fp) );
      adbase = address << 4;
      break;
    case 0x03:  /* no meaning(start address) */
      address = hex( getc(fp) );
      address = address * 16 + hex( getc(fp) );
      address = address * 16 + hex( getc(fp) );
      address = address * 16 + hex( getc(fp) );
      address = address * 16 + hex( getc(fp) );
      address = address * 16 + hex( getc(fp) );
      address = address * 16 + hex( getc(fp) );
      address = address * 16 + hex( getc(fp) );
      printf("start address=%08x\n",((address>>16)<<4)+(address&0xffff));
      break;
    case 0x04:  /* upper 16bit for 32bit linear addr */
      address = hex( getc(fp) );
      address = address * 16 + hex( getc(fp) );
      address = address * 16 + hex( getc(fp) );
      address = address * 16 + hex( getc(fp) );
      adbase = address << 16;
      break;
    case 0x05:  /* no meaning(start address for 32bit linear */
      address = hex( getc(fp) );
      address = address * 16 + hex( getc(fp) );
      address = address * 16 + hex( getc(fp) );
      address = address * 16 + hex( getc(fp) );
      address = address * 16 + hex( getc(fp) );
      address = address * 16 + hex( getc(fp) );
      address = address * 16 + hex( getc(fp) );
      address = address * 16 + hex( getc(fp) );
      printf("start address=%08x\n",address);
      break;
    case 0x32:  /* extended 32-bit address code(original) */
      address = hex( getc(fp) );
      address = address * 16 + hex( getc(fp) );
      address = address * 16 + hex( getc(fp) );
      address = address * 16 + hex( getc(fp) );
      address = address * 16 + hex( getc(fp) );
      address = address * 16 + hex( getc(fp) );
      address = address * 16 + hex( getc(fp) );
      address = address * 16 + hex( getc(fp) );
      adbase = address;
      break;
    }	
  }
}

/*
  Motorola S Record Load

S0
     Header record for each block of S records which may contain descriptive
     information identifying the following block of S records. The address
     field is typically zeros. The code/data field of an S0 record typically
     contains a "hexified" ASCII string.
S1
     Code or data record containing a 2-byte address (4 hex digit address).
S2
     Code or data record containing a 3-byte address (6 hex digit address).
S3
     Code or data record containing a 4-byte address (8 hex digit address).
S5
     Count record containing the number of S1, S2, and S3 records
     transmitted in a block. The count appears in the address field, and
     there is no code/data field (I never saw this actually used).
S7
     Termination record for a block of S records. A 4-byte address contains
     the address where execution starts. There is no code/data field.
S8
     Termination record for a block of S records. A 3-byte address contains
     the address where execution starts. There is no code/data field.
S9
     Termination record for a block of S records. A 2-byte address contains
     the address where execution starts. There is no code or data field.
*/
void loads(char *fname, char *base)
{

  int eof, i;
  int ch;

  int	reclen;
  int   addlen;
  long  adbase, address;
  int	byte;
  int   load;

  adbase = 0;
  fp = fopen(fname,"r");
  if(fp==NULL){
    fprintf(stderr,"cannot open %s\n",fname);
    return;
  }
  eof = 0;
  while(!eof){
    while((ch=getc(fp)) != 'S')
      if(ch==EOF){
	fclose(fp);
	return;
      }
    switch( getc(fp) ){
    case '1':
      addlen = 2;
      load   = 1;
      break;
    case '2':
      addlen = 3;
      load   = 1;
      break;
    case '3':
      addlen = 4;
      load   = 1;
      break;
    case '7':
    case '8':
    case '9':
      reclen = hex( getc(fp) );
      reclen = reclen * 16 + hex( getc(fp) );
      address = hex( getc(fp) );
      address = address * 16 + hex( getc(fp) );
      address = address * 16 + hex( getc(fp) );
      address = address * 16 + hex( getc(fp) );
      if(addlen>2){
	address = address * 16 + hex( getc(fp) );
	address = address * 16 + hex( getc(fp) );
	if(addlen>3){
	  address = address * 16 + hex( getc(fp) );
	  address = address * 16 + hex( getc(fp) );
	}
      }
      printf("start address=0x%08x\n",address);
      eof = 1;
      load = 0;
      fclose(fp);
      break;
    default:
      load = 0;
    }
    if(load){
      reclen = hex( getc(fp) );
      reclen = reclen * 16 + hex( getc(fp) );
      address = hex( getc(fp) );
      address = address * 16 + hex( getc(fp) );
      address = address * 16 + hex( getc(fp) );
      address = address * 16 + hex( getc(fp) );
      if(addlen>2){
	address = address * 16 + hex( getc(fp) );
	address = address * 16 + hex( getc(fp) );
	if(addlen>3){
	  address = address * 16 + hex( getc(fp) );
	  address = address * 16 + hex( getc(fp) );
	}
      }
      address += adbase;
      for(i=0; i<(reclen-addlen); i++){
	byte = hex( getc(fp) );
	byte = byte * 16 + hex( getc(fp) );
	put_byte(address+i, byte);
      }
    }
  }
}

static char Line[1024];

void loadm(char *fname)
{
  int   eof;
  int addr, data;

  fp = fopen(fname,"r");
  if(fp==NULL){
    fprintf(stderr,"cannot open %s\n",fname);
    return;
  }
  eof = 0;
  while(!eof){
    if(fgets(Line, 1024, fp)==0){
      fclose(fp);
      eof = 1;
      continue;
    }
    if(strlen(Line)>16){ /* MIPS32 */
      sscanf(Line, "%x %x", &addr, &data);
      put_word(addr, data);
    }
    else{/* MIPS16 */
      sscanf(Line, "%x %x", &addr, &data);
      put_byte(addr+0, data&0xffL);
      put_byte(addr+1, (data>>8)&0xffL);
    }
   
  }
}

long get_dat(FILE* fp, int size)
{
  int i;
  int n;
  char ch[16];
  if(size>16)
    size = 16;
  for(i=0;i<size;i++)
    ch[i] = fgetc(fp);
  switch(size){
  case 1:
    return ch[0];
  case 2:
    return (ch[0]<<8)|(ch[1]&0xff);
  default:
    return (ch[0]<<24)|((ch[1]&0xff)<<16)|((ch[2]&0xff)<<8)|(ch[3]&0xff);
  }
}

void loade(char *fname)
{
  int arch_64;
  int endi_big;

/* ELF Header */
  int e_type;
  int e_machine;
  int e_version;
  int e_entry;
  int e_phoff;
  int e_shoff;
  int e_flags;
  int e_ehsize;
  int e_phentsize;
  int e_phnum;
  int e_shentsize;
  int e_shnum;
  int e_shstrndx;
/* Program Header */
  int p_type;
  int p_offset;
  int p_vaddr;
  int p_paddr;
  int p_filesz;
  int p_memsz;
  int p_flags;
  int p_align;
  int m0,m1,m2,m3;
  int ei_class;
  int ei_data;
  int i;
  int ph_off;

  fp = fopen(fname,"rb");
  if(fp==NULL){
    fprintf(stderr,"cannot open %s\n",fname);
    return;
  }
  m0 = get_dat(fp,1);
  m1 = get_dat(fp,1);
  m2 = get_dat(fp,1);
  m3 = get_dat(fp,1);
  ei_class = get_dat(fp,1);
  ei_data  = get_dat(fp,1);
  if(m0!=0x7f){
    fprintf(stderr,"not ELF file %s!\n",fname);
    fclose(fp);
    return;
  }
  if(m1!='E'){
    fprintf(stderr,"not ELF file %s!\n",fname);
    fclose(fp);
    return;
  }
  if(m2!='L'){
    fprintf(stderr,"not ELF file %s!\n",fname);
    fclose(fp);
    return;
  }
  if(m3!='F'){
    fprintf(stderr,"not ELF file %s!\n",fname);
    fclose(fp);
    return;
  }
  if(ei_class == 2)
    arch_64 = 1;
  else
    arch_64 = 0;
  if(ei_data == 2)
    endi_big = 1;
  else
    endi_big = 0;
/*
  printf("arch_64=%d endi_big=%d\n",arch_64, endi_big);
*/
  fseek(fp, 16, SEEK_SET);
  if(arch_64){
    e_type      = get_dat(fp,2);
    e_machine   = get_dat(fp,2);
    e_version   = get_dat(fp,4);
    e_entry     = get_dat(fp,4);
    fseek(fp, 4, SEEK_CUR);
    e_phoff     = get_dat(fp,4);
    fseek(fp, 4, SEEK_CUR);
    e_shoff     = get_dat(fp,4);
    fseek(fp, 4, SEEK_CUR);
    e_flags     = get_dat(fp,4);
    e_ehsize    = get_dat(fp,2);
    e_phentsize = get_dat(fp,2);
    e_phnum     = get_dat(fp,2);
    e_shentsize = get_dat(fp,2);
    e_shnum     = get_dat(fp,2);
    e_shstrndx  = get_dat(fp,2);
  }
  else{
    e_type      = get_dat(fp,2);
    e_machine   = get_dat(fp,2);
    e_version   = get_dat(fp,4);
    e_entry     = get_dat(fp,4);
    e_phoff     = get_dat(fp,4);
    e_shoff     = get_dat(fp,4);
    e_flags     = get_dat(fp,4);
    e_ehsize    = get_dat(fp,2);
    e_phentsize = get_dat(fp,2);
    e_phnum     = get_dat(fp,2);
    e_shentsize = get_dat(fp,2);
    e_shnum     = get_dat(fp,2);
    e_shstrndx  = get_dat(fp,2);
  }
  printf("start address=0x%08x\n",e_entry);
#if 0
  printf("arch_64=%x\n",arch_64);
  printf("endi_big=%x\n",endi_big);
  printf("e_type=%x\n",e_type);
  printf("e_machine=%x\n",e_machine);
  printf("e_entry=%x\n",e_entry);
  printf("e_phoff=%x\n",e_phoff);
  printf("e_shoff=%x\n",e_shoff);
  printf("e_flags=%x\n",e_flags);
  printf("e_ehsize=%x\n",e_ehsize);
  printf("e_phentsize=%x\n",e_phentsize);
  printf("e_phnum=%x\n",e_phnum);
  printf("e_shentsize=%x\n",e_shentsize);
  printf("e_shnum=%x\n",e_shnum);
  printf("e_shstrndx=%x\n",e_shstrndx);
#endif
  ph_off = 0;
  for(i=0;i<e_phnum;i++){
    fseek(fp, e_phoff+ph_off, SEEK_SET);
    if(arch_64){
      p_type   = get_dat(fp,4);
      p_flags  = get_dat(fp,4);
      p_offset = get_dat(fp,4);
      fseek(fp, 4, SEEK_CUR);
      p_vaddr  = get_dat(fp,4);
      fseek(fp, 4, SEEK_CUR);
      p_paddr  = get_dat(fp,4);
      fseek(fp, 4, SEEK_CUR);
      p_filesz = get_dat(fp,4);
      fseek(fp, 4, SEEK_CUR);
      p_memsz  = get_dat(fp,4);
      fseek(fp, 4, SEEK_CUR);
      p_align  = get_dat(fp,4);
      fseek(fp, 4, SEEK_CUR);
      ph_off += 56;
    }
    else{
      p_type   = get_dat(fp,4);
      p_offset = get_dat(fp,4);
      p_vaddr  = get_dat(fp,4);
      p_paddr  = get_dat(fp,4);
      p_filesz = get_dat(fp,4);
      p_memsz  = get_dat(fp,4);
      p_flags  = get_dat(fp,4);
      p_align  = get_dat(fp,4);
      ph_off += 32;
    }
    //if((p_vaddr&0xc0000000)==0x80000000){
    //  p_paddr = p_vaddr & 0x1fffffff;
    //}
#if 0
    printf("p_type=%x\n",p_type);
    printf("p_offset=%x\n",p_offset);
    printf("p_vaddr=%x\n",p_vaddr);
    printf("p_paddr=%x\n",p_paddr);
    printf("p_filesz=%x\n",p_filesz);
    printf("p_memsz=%x\n",p_memsz);
    printf("p_flags=%x\n",p_flags);
    printf("p_align=%x\n",p_align);
#endif
    fseek(fp, p_offset, SEEK_SET);
    while(p_memsz){
#if 0
      printf("p_paddr=%08x p_memsz=%x\n",p_paddr,p_memsz);
#endif
      put_byte(p_vaddr, fgetc(fp));
      p_vaddr++;
      p_memsz--;
    }
  }
  fclose(fp);
}

  
