#include <stdio.h>
#include "define.h"
/*
  Memory Module
*/
static char mem[MAXPAGENUM][0x10000];
static int pair[MAXPAGENUM];
static int pair_ptr= -1;

static int search(int addr)
{
  int i;

  if(pair_ptr<0){ /* init */
    pair_ptr = 0;
    pair[pair_ptr] = addr&0xffff0000L;
    return pair_ptr;
  }
  for(i=0;i<=pair_ptr;i++){
    if(((pair[i]^addr)&0xffff0000L)==0)
      break;
  }
  if(i<=pair_ptr){ /* match */
    return i;
  }
  if((pair_ptr+1)<MAXPAGENUM){ /* allocate */
    pair_ptr++;
    pair[pair_ptr] = addr&0xffff0000L;
    return pair_ptr;
  }
  printf("mem.c:out of memory at 0x%08x\n",addr);
  return -1;
}

int get_word(int addr)
{
  int i;
  int c0,c1,c2,c3;

  addr &= 0xfffffffcL;
  i = search(addr);
  if(i>=0){
    addr &= 0xffff;
    c0 = mem[i][addr+0]&0xff;
    c1 = mem[i][addr+1]&0xff;
    c2 = mem[i][addr+2]&0xff;
    c3 = mem[i][addr+3]&0xff;
#if 0
    printf("GET %x%04x %08x\n",i,addr, (c3<<24)|(c2<<16)|(c1<<8)|(c0));
#endif
    return (c3<<24)|(c2<<16)|(c1<<8)|(c0);
  }
  return 0;
}

void put_word(int addr, int data)
{
  int i;
  int c0,c1,c2,c3;

  addr &= 0xfffffffcL;
  i = search(addr);
  if(i>=0){
    addr &= 0xffff;
    c0 = (data    )&0xff;
    c1 = (data>> 8)&0xff;
    c2 = (data>>16)&0xff;
    c3 = (data>>24)&0xff;
    mem[i][addr+0] = c0;
    mem[i][addr+1] = c1;
    mem[i][addr+2] = c2;
    mem[i][addr+3] = c3;
#if 0
    printf("PUT %x%04x %08x\n",i,addr, (c3<<24)|(c2<<16)|(c1<<8)|(c0));
#endif
  }
}

static int swap_w(int d)
{
  int hh,hl,lh,ll;
  hh = (d>>24)&0xff;
  hl = (d>>16)&0xff;
  lh = (d>> 8)&0xff;
  ll = (d>> 0)&0xff;
  return (ll<<24)|(lh<<16)|(hl<<8)|hh;
}

static int swap_h(int d)
{
  int h,l;
  h = (d>>8)&0xff;
  l = (d>>0)&0xff;
  return (l<<8)|h;
}

int get_mem(int addr, int size)
{
  int d;

  d = get_word(addr);

  if(size==WORD)
    return swap_w(d);
  if(size==HWORD){
    if(addr&2)/* upper */
      return swap_h((d>>16)&0xffff);
    else /* lower */
      return swap_h(d&0xffff);
  }
  if(size==BYTE){
    switch(addr&3){
    case 0:
      return d&0xff;
    case 1:
      return (d>>8)&0xff;
    case 2:
      return (d>>16)&0xff;
    case 3:
      return (d>>24)&0xff;
    }
  }
  return 0;
}

void put_mem(int addr, int value, int size)
{
  int d;

  if(size==WORD){
    put_word(addr, swap_w(value));
    return;
  }
  d = get_word(addr);
  if(size==HWORD){
    if(addr&2){/* upper */
      put_word(addr, ((swap_h(value)&0xffff)<<16)|(d&0xffff));
      return;
    }
    else {/* lower */
      put_word(addr, (swap_h(value)&0xffff)|(d&0xffff0000));
      return;
    }
  }
  if(size==BYTE){
    switch(addr&3){
    case 0:
      put_word(addr, (value&0x000000ff)|(d&0xffffff00));
      return;
    case 1:
      put_word(addr, ((value<<8)&0x0000ff00)|(d&0xffff00ff));
      return;
    case 2:
      put_word(addr, ((value<<16)&0x00ff0000)|(d&0xff00ffff));
      return;
    case 3:
      put_word(addr, ((value<<24)&0xff000000)|(d&0x00ffffff));
      return;
    }
  }
}


void mem_dump(unsigned int start, unsigned int end, int unit)
{
  unsigned int  m_start;
  unsigned int  m_end;
  unsigned int  a,c;
  int    width,ainc;
  int    data;
  unsigned char   asc[16];
  unsigned char   d0,d1,d2,d3;

  if(unit==0){
    width=16;
    ainc =1;
  }
  else if(unit==1){
    width=8;
    ainc =2;
  }
  else {
    unit =2;
    width=4;
    ainc =4;
  }
  m_start = start&0xfffffffcL;
  m_end   = end  &0xfffffffcL;
  for(a=m_start,c=0;a<m_end;a += 4){
    if(c==0) printf("%08x: ",a);
    data=get_word(a);
    asc[4*c+0]=d0=data&0xff;
    asc[4*c+1]=d1=(data>>8)&0xff;
    asc[4*c+2]=d2=(data>>16)&0xff;
    asc[4*c+3]=d3=(data>>24)&0xff;
    switch(unit){
    case 0:
      printf("%02x %02x %02x %02x  ",d0,d1,d2,d3);
      break;
    case 1:
      printf("%04x %04x ",(d0<<8)|d1,(d2<<8)|d3);
      break;
    case 2:
      printf("%08x ",swap_w(data));
      break;
    }
    asc[4*c+0]=((d0>=' ')&&(d0<='~')) ? d0 : '.';
    asc[4*c+1]=((d1>=' ')&&(d1<='~')) ? d1 : '.';
    asc[4*c+2]=((d2>=' ')&&(d2<='~')) ? d2 : '.';
    asc[4*c+3]=((d3>=' ')&&(d3<='~')) ? d3 : '.';
    c++;
    if(c==4){
      int     i;
      c=0;
      for(i=0;i<16;i++) printf("%c",asc[i]);
      printf("\n");
    }
  }
  if(c!=0){
    int i;
    while(c<4){
      switch(unit){
      case 0:
	printf("            ",d0,d1,d2,d3);
	break;
      case 1:
	printf("          ",(d1<<8)|d0,(d3<<8)|d2);
	break;
      case 2:
	printf("         ",data);
	break;
      }
      asc[4*c+0]=' ';
      asc[4*c+1]=' ';
      asc[4*c+2]=' ';
      asc[4*c+3]=' ';
      c++;
    }
    for(i=0;i<16;i++) printf("%c",asc[i]);
    printf("\n");
  }
}
